from __future__ import annotations

from collections.abc import Iterable
from typing import Any

import msgspec

from ..persistence.save_status import GameStatusData
from ..replay.checkpoints import ReplayCheckpoint
from .canonical_channels import EntitySamplesSnapshot, RngStreamRow, SimStateSnapshot, TimingSampleRow

TRACE_MAGIC = b"crimson_debug_trace_v1\n"
TRACE_FORMAT_VERSION = 1
TRACE_SCHEMA_VERSION = 10
SUPPORTED_TRACE_SCHEMA_VERSIONS = frozenset((TRACE_SCHEMA_VERSION,))

TRACE_REQUIRED_CHANNELS = (
    "checkpoint",
    "sim_state",
    "entity_samples",
    "rng_stream",
    "timing_samples",
)

_DEFAULT_CHANNEL_VERSION = 1

_CHUNK_KIND_META = b"META"
_CHUNK_KIND_TICK = b"TICK"
_CHUNK_KIND_FOOTER = b"FOTR"
CHUNK_KINDS = frozenset((_CHUNK_KIND_META, _CHUNK_KIND_TICK, _CHUNK_KIND_FOOTER))

CHUNK_KIND_META = _CHUNK_KIND_META.decode("ascii")
CHUNK_KIND_TICK = _CHUNK_KIND_TICK.decode("ascii")
CHUNK_KIND_FOOTER = _CHUNK_KIND_FOOTER.decode("ascii")

TRAILER_MAGIC = b"CDTFTR1\n"

CHUNK_FLAG_ZSTD = 1 << 0
CHUNK_FLAG_MSGPACK = 1 << 1
DEFAULT_CHUNK_FLAGS = int(CHUNK_FLAG_ZSTD | CHUNK_FLAG_MSGPACK)


class TraceMeta(msgspec.Struct):
    trace_format_version: int
    trace_schema_version: int
    created_utc: str
    producer: dict[str, Any]
    source: dict[str, Any]
    channels: list[str]
    channel_versions: dict[str, int]
    tick_range: dict[str, int]
    config: dict[str, Any]
    status: GameStatusData | None = None


class ReplayTickChannels(msgspec.Struct):
    checkpoint: ReplayCheckpoint
    sim_state: SimStateSnapshot
    entity_samples: EntitySamplesSnapshot
    rng_stream: list[RngStreamRow] = msgspec.field(default_factory=list)
    timing_samples: list[TimingSampleRow] = msgspec.field(default_factory=list)


class TickRecord(msgspec.Struct):
    tick_index: int
    elapsed_ms: int
    dt_ms_i32: int
    mode_id: int
    channels: ReplayTickChannels
    phase_markers: list[str] = msgspec.field(default_factory=list)


class TickBlock(msgspec.Struct):
    start_tick: int
    end_tick: int
    ticks: list[TickRecord]


class TickBlockIndexEntry(msgspec.Struct):
    start_tick: int
    end_tick: int
    file_offset: int
    compressed_len: int
    uncompressed_len: int
    checksum: int


class TraceFooter(msgspec.Struct):
    trace_format_version: int
    tick_blocks: list[TickBlockIndexEntry]
    tick_count: int
    first_tick: int
    last_tick: int
    channel_counts: dict[str, int]


def channel_versions_for(channels: Iterable[str]) -> dict[str, int]:
    return {str(channel): int(_DEFAULT_CHANNEL_VERSION) for channel in channels}



from __future__ import annotations

from ..game_modes import GameMode
from ..quests.level import QuestLevel
from ..quests.types import SpawnEntry
from ..sim.world_state import WorldState
from ..tutorial import reset_tutorial_state
from ..tutorial.runtime import tutorial_before_step, tutorial_input_transform, tutorial_post_step
from ..typo.runtime import typo_before_step, typo_input_transform, typo_mid_step, typo_post_step
from ..typo.state import reset_typo_state
from ..weapon_runtime import weapon_assign_player
from ..weapons import WeaponId
from .sessions import (
    DeterministicSession,
    QuestSpawnState,
    RushSpawnState,
    SurvivalSpawnState,
    quest_post_step,
    rush_input_transform,
    rush_mid_step,
    survival_mid_step,
)

RUSH_WEAPON_ID = WeaponId.ASSAULT_RIFLE
RUSH_FORCED_AMMO = 30.0


def enforce_rush_loadout(world: WorldState) -> None:
    for player in world.players:
        if player.weapon.weapon_id != RUSH_WEAPON_ID:
            weapon_assign_player(player, RUSH_WEAPON_ID, state=world.state)
        # Native `rush_mode_update` forces assault rifle + 30 ammo every frame.
        player.weapon.ammo = float(RUSH_FORCED_AMMO)


def build_survival_session(
    *,
    world: WorldState,
    world_size: float,
    damage_scale_by_type: dict[int, float],
    detail_preset: int,
    violence_disabled: int,
    game_tune_started: bool,
    finalize_post_render_lifecycle: bool,
    apply_world_dt_steps: bool = False,
) -> tuple[DeterministicSession, SurvivalSpawnState]:
    spawn_state = SurvivalSpawnState()
    session = DeterministicSession(
        world=world,
        world_size=world_size,
        damage_scale_by_type=damage_scale_by_type,
        game_mode=GameMode.SURVIVAL,
        perk_progression_enabled=True,
        detail_preset=detail_preset,
        violence_disabled=violence_disabled,
        game_tune_started=game_tune_started,
        apply_world_dt_steps=apply_world_dt_steps,
        finalize_post_render_lifecycle=finalize_post_render_lifecycle,
        mid_step_hook=lambda ctx: survival_mid_step(ctx, spawn_state),
    )
    return session, spawn_state


def build_rush_session(
    *,
    world: WorldState,
    world_size: float,
    damage_scale_by_type: dict[int, float],
    detail_preset: int,
    violence_disabled: int,
    game_tune_started: bool,
    finalize_post_render_lifecycle: bool,
) -> tuple[DeterministicSession, RushSpawnState]:
    spawn_state = RushSpawnState()
    session = DeterministicSession(
        world=world,
        world_size=world_size,
        damage_scale_by_type=damage_scale_by_type,
        game_mode=GameMode.RUSH,
        perk_progression_enabled=False,
        detail_preset=detail_preset,
        violence_disabled=violence_disabled,
        game_tune_started=game_tune_started,
        finalize_post_render_lifecycle=finalize_post_render_lifecycle,
        elapsed_uses_raw_dt=True,
        mid_step_hook=lambda ctx: rush_mid_step(ctx, spawn_state),
        before_step_hook=lambda: enforce_rush_loadout(world),
        input_transform=rush_input_transform,
    )
    return session, spawn_state


def build_quest_session(
    *,
    world: WorldState,
    world_size: float,
    damage_scale_by_type: dict[int, float],
    detail_preset: int,
    violence_disabled: int,
    game_tune_started: bool,
    demo_mode_active: bool,
    apply_world_dt_steps: bool,
    finalize_post_render_lifecycle: bool,
    spawn_entries: tuple[SpawnEntry, ...],
    quest_level: QuestLevel | None,
    start_weapon_id: WeaponId | None,
) -> tuple[DeterministicSession, QuestSpawnState]:
    world.state.quest_level = quest_level

    weapon_id = WeaponId.PISTOL if start_weapon_id in (None, WeaponId.NONE) else start_weapon_id
    for player in world.players:
        weapon_assign_player(player, weapon_id, state=world.state)

    world.creatures.capture_spawn_events_authoritative = False
    quest_state = QuestSpawnState(spawn_entries=tuple(spawn_entries))
    session = DeterministicSession(
        world=world,
        world_size=world_size,
        damage_scale_by_type=damage_scale_by_type,
        game_mode=GameMode.QUESTS,
        perk_progression_enabled=True,
        detail_preset=detail_preset,
        violence_disabled=violence_disabled,
        game_tune_started=game_tune_started,
        demo_mode_active=demo_mode_active,
        apply_world_dt_steps=apply_world_dt_steps,
        finalize_post_render_lifecycle=finalize_post_render_lifecycle,
        post_step_hook=lambda ctx: quest_post_step(ctx, quest_state),
    )
    return session, quest_state


def build_typo_session(
    *,
    world: WorldState,
    world_size: float,
    damage_scale_by_type: dict[int, float],
    detail_preset: int,
    violence_disabled: int,
    game_tune_started: bool,
    dictionary_words: tuple[str, ...] = (),
    highscore_names: tuple[str, ...] = (),
) -> DeterministicSession:
    reset_typo_state(
        world.state.typo,
        creature_capacity=len(world.creatures.entries),
        dictionary_words=dictionary_words,
        highscore_names=highscore_names,
    )
    session = DeterministicSession(
        world=world,
        world_size=world_size,
        damage_scale_by_type=damage_scale_by_type,
        game_mode=GameMode.TYPO,
        perk_progression_enabled=False,
        detail_preset=detail_preset,
        violence_disabled=violence_disabled,
        game_tune_started=game_tune_started,
        before_step_hook=lambda: typo_before_step(world),
        mid_step_hook=typo_mid_step,
        post_step_hook=typo_post_step,
        input_transform=lambda inputs: typo_input_transform(world, inputs),
    )
    return session


def build_tutorial_session(
    *,
    world: WorldState,
    world_size: float,
    damage_scale_by_type: dict[int, float],
    detail_preset: int,
    violence_disabled: int,
    game_tune_started: bool,
    demo_mode_active: bool,
) -> DeterministicSession:
    reset_tutorial_state(
        world.state.tutorial,
        world.state.tutorial_overlay,
        preserve_bugs=world.state.preserve_bugs,
    )
    session = DeterministicSession(
        world=world,
        world_size=world_size,
        damage_scale_by_type=damage_scale_by_type,
        game_mode=GameMode.TUTORIAL,
        perk_progression_enabled=True,
        detail_preset=detail_preset,
        violence_disabled=violence_disabled,
        game_tune_started=game_tune_started,
        demo_mode_active=demo_mode_active,
        before_step_hook=lambda: tutorial_before_step(world),
        post_step_hook=tutorial_post_step,
        input_transform=lambda inputs: tutorial_input_transform(world, inputs),
    )
    return session



from __future__ import annotations

from .codec import ReplayCodecError, dump_replay, dump_replay_file, load_replay, load_replay_file
from .input_codec import pack_player_input, pack_tick_inputs, unpack_player_input, unpack_tick_inputs
from .recorder import ReplayRecorder
from .types import (
    AIM_SCHEME_PRESENT_FLAG,
    AIM_SCHEME_SHIFT,
    FIRE_DOWN_FLAG,
    FIRE_PRESSED_FLAG,
    MOVE_BACKWARD_FLAG,
    MOVE_FORWARD_FLAG,
    MOVE_KEYS_PRESENT_FLAG,
    MOVE_MODE_PRESENT_FLAG,
    MOVE_MODE_SHIFT,
    RELOAD_PRESSED_FLAG,
    TURN_LEFT_FLAG,
    TURN_RIGHT_FLAG,
    PackedPlayerInput,
    PackedTickInputs,
    Replay,
    ReplayClaimedStatsSnapshot,
    ReplayHeader,
    ReplayTick,
    pack_input_flags,
    unpack_input_flags,
    unpack_input_mode_flags,
    unpack_input_move_key_flags,
    unpack_packed_player_input,
)
from .versioning import ReplayGameVersionError, ReplayGameVersionWarning, warn_on_game_version_mismatch

__all__ = [
    "FIRE_DOWN_FLAG",
    "FIRE_PRESSED_FLAG",
    "AIM_SCHEME_PRESENT_FLAG",
    "AIM_SCHEME_SHIFT",
    "MOVE_BACKWARD_FLAG",
    "MOVE_FORWARD_FLAG",
    "MOVE_MODE_PRESENT_FLAG",
    "MOVE_MODE_SHIFT",
    "MOVE_KEYS_PRESENT_FLAG",
    "RELOAD_PRESSED_FLAG",
    "TURN_LEFT_FLAG",
    "TURN_RIGHT_FLAG",
    "PackedPlayerInput",
    "PackedTickInputs",
    "ReplayClaimedStatsSnapshot",
    "Replay",
    "ReplayCodecError",
    "ReplayHeader",
    "ReplayRecorder",
    "ReplayTick",
    "ReplayGameVersionError",
    "ReplayGameVersionWarning",
    "dump_replay",
    "dump_replay_file",
    "load_replay",
    "load_replay_file",
    "pack_player_input",
    "pack_input_flags",
    "pack_tick_inputs",
    "unpack_player_input",
    "unpack_packed_player_input",
    "unpack_input_flags",
    "unpack_input_mode_flags",
    "unpack_input_move_key_flags",
    "unpack_tick_inputs",
    "warn_on_game_version_mismatch",
]



from __future__ import annotations

import msgspec

from grim.sfx_map import SfxId

from ..math_parity import f32
from .presentation_step import PresentationStepCommands
from .terrain_fx import TerrainFxBatch
from .timing import FrameTiming
from .world_state import WorldEvents


class PresentationRngTrace(msgspec.Struct):
    draws_total: int = 0


class DeterministicStepResult(msgspec.Struct):
    dt_sim: float
    timing: FrameTiming
    events: WorldEvents
    presentation: PresentationStepCommands
    presentation_plan_ms: float
    presentation_rng_trace: PresentationRngTrace
    terrain_fx: TerrainFxBatch = TerrainFxBatch()
    post_apply_sfx: tuple[SfxId, ...] = ()


def time_scale_reflex_boost_bonus(
    *,
    reflex_boost_timer: float,
    time_scale_active: bool,
    dt: float,
) -> float:
    """Apply Reflex Boost time scaling, matching the classic frame loop latch semantics."""

    # Native stores frame delta time in float32 (`frame_dt`). Many downstream systems
    # multiply `frame_dt` before rounding back to float32, so the *input* precision
    # matters even when Reflex Boost is inactive.
    dt_f32 = f32(float(dt))
    if not (float(dt_f32) > 0.0):
        return float(dt_f32)
    if not time_scale_active:
        return float(dt_f32)

    time_scale_factor = time_scale_reflex_boost_factor(
        reflex_boost_timer=float(reflex_boost_timer),
        time_scale_active=bool(time_scale_active),
    )
    return float(f32(float(dt_f32) * float(time_scale_factor)))


def time_scale_reflex_boost_factor(
    *,
    reflex_boost_timer: float,
    time_scale_active: bool,
) -> float:
    if not bool(time_scale_active):
        return 1.0
    reflex_f32 = f32(float(reflex_boost_timer))
    time_scale_factor = f32(0.3)
    if float(reflex_f32) < 1.0:
        time_scale_factor = f32((1.0 - float(reflex_f32)) * 0.7 + 0.3)
    return float(time_scale_factor)



from __future__ import annotations

import math
from typing import cast

import msgspec

from grim.geom import Vec2

from ..creatures.runtime import CreaturePool
from ..creatures.spawn import SpawnEnv
from ..gameplay import GameplayState
from ..sim.presentation_step import PresentationStepCommands
from ..sim.state_types import PlayerState
from ..sim.world_state import WorldEvents, WorldState
from ..weapon_runtime import init_default_alt_weapon, weapon_assign_player
from ..weapons import WEAPON_TABLE, WeaponId


def _weapon_damage_scale_map() -> dict[int, float]:
    table: dict[int, float] = {}
    for entry in WEAPON_TABLE:
        if int(entry.weapon_id) <= 0:
            continue
        table[int(entry.weapon_id)] = float(cast(float, entry.damage_scale))
    return table


def reset_world_players(
    players: list[PlayerState],
    *,
    state: GameplayState,
    world_size: float,
    player_count: int,
    spawn_pos: Vec2 | None = None,
) -> None:
    players.clear()

    base = Vec2(float(world_size) * 0.5, float(world_size) * 0.5) if spawn_pos is None else spawn_pos
    count = max(1, int(player_count))
    if count <= 1:
        offsets = [Vec2()]
    else:
        radius = 32.0
        step = math.tau / float(count)
        offsets = [Vec2.from_angle(float(idx) * step) * radius for idx in range(count)]

    for idx in range(count):
        pos = (base + offsets[idx]).clamp_rect(0.0, 0.0, float(world_size), float(world_size))
        player = PlayerState(index=idx, pos=pos)
        weapon_assign_player(player, WeaponId.PISTOL, state=state)
        init_default_alt_weapon(player)
        players.append(player)

    # Reset-time loadout bootstrap should not leak queued reload SFX.
    state.sfx_queue.clear()


class SimWorldState(msgspec.Struct):
    world_size: float = 1024.0
    demo_mode_active: bool = False
    quest_fail_retry_count: int = 0
    hardcore: bool = False
    preserve_bugs: bool = False

    world_state: WorldState = cast(WorldState, None)
    spawn_env: SpawnEnv = cast(SpawnEnv, None)
    state: GameplayState = cast(GameplayState, None)
    players: list[PlayerState] = msgspec.field(default_factory=list)
    creatures: CreaturePool = cast(CreaturePool, None)

    damage_scale_by_type: dict[int, float] = msgspec.field(default_factory=_weapon_damage_scale_map)
    presentation_elapsed_ms: float = 0.0
    bonus_anim_phase: float = 0.0
    game_tune_started: bool = False
    last_events: WorldEvents = msgspec.field(
        default_factory=lambda: WorldEvents(hits=[], deaths=(), pickups=[], sfx=[]),
    )
    last_presentation: PresentationStepCommands = msgspec.field(default_factory=PresentationStepCommands)
    def __post_init__(self) -> None:
        self.reset(seed=0xBEEF, player_count=1)

    def reset(
        self,
        *,
        seed: int = 0xBEEF,
        player_count: int = 1,
        spawn_pos: Vec2 | None = None,
    ) -> None:
        self.world_state = WorldState.build(
            world_size=float(self.world_size),
            demo_mode_active=bool(self.demo_mode_active),
            hardcore=bool(self.hardcore),
            quest_fail_retry_count=int(self.quest_fail_retry_count),
            preserve_bugs=bool(self.preserve_bugs),
        )
        self.spawn_env = self.world_state.spawn_env
        self.state = self.world_state.state
        self.players = self.world_state.players
        self.creatures = self.world_state.creatures
        self.state.rng.srand(int(seed))

        self.last_events = WorldEvents(hits=[], deaths=(), pickups=[], sfx=[])
        self.last_presentation = PresentationStepCommands()

        self.presentation_elapsed_ms = 0.0
        self.bonus_anim_phase = 0.0
        self.game_tune_started = False

        reset_world_players(
            self.players,
            state=self.state,
            world_size=float(self.world_size),
            player_count=int(player_count),
            spawn_pos=spawn_pos,
        )

    def load_world_state(self, world_state: WorldState) -> None:
        self.world_state = world_state
        self.spawn_env = self.world_state.spawn_env
        self.state = self.world_state.state
        self.players = self.world_state.players
        self.creatures = self.world_state.creatures
        self.last_events = WorldEvents(hits=[], deaths=(), pickups=[], sfx=[])
        self.last_presentation = PresentationStepCommands()


    def apply_step_metadata(
        self,
        *,
        events: WorldEvents,
        presentation: PresentationStepCommands,
        dt_sim: float,
        game_tune_started: bool,
    ) -> None:
        self.last_events = events
        self.last_presentation = presentation

        if float(dt_sim) > 0.0:
            self.presentation_elapsed_ms += float(dt_sim) * 1000.0
            self.bonus_anim_phase += float(dt_sim) * 1.3

        self.game_tune_started = bool(game_tune_started)

    def close_session(self) -> None:
        self.last_events = WorldEvents(hits=[], deaths=(), pickups=[], sfx=[])
        self.last_presentation = PresentationStepCommands()

        self.game_tune_started = False



from __future__ import annotations

from collections.abc import Callable
from typing import cast

import msgspec

from grim.audio import AudioState
from grim.rand import Crand

from ..audio_router import AudioRouter
from ..sim.presentation_step import PresentationStepCommands, apply_presentation_plan


def _zero_reflex_boost() -> float:
    return 0.0


class AudioBridge(msgspec.Struct):
    audio_rng: Crand
    demo_mode_active: bool = False
    reflex_boost_timer_source: Callable[[], float] = _zero_reflex_boost
    audio: AudioState | None = None
    router: AudioRouter = cast(AudioRouter, None)

    def __post_init__(self) -> None:
        self.router = AudioRouter(
            audio_rng=self.audio_rng,
            audio=self.audio,
            demo_mode_active=bool(self.demo_mode_active),
            reflex_boost_timer_source=self.reflex_boost_timer_source,
        )

    def sync(
        self,
        *,
        audio: AudioState | None,
        audio_rng: Crand,
        demo_mode_active: bool,
    ) -> None:
        self.audio = audio
        self.audio_rng = audio_rng
        self.demo_mode_active = bool(demo_mode_active)
        self.router.audio = audio
        self.router.audio_rng = audio_rng
        self.router.demo_mode_active = bool(demo_mode_active)

    def apply_plan(self, *, plan: PresentationStepCommands, apply_audio: bool = True) -> None:
        apply_presentation_plan(
            plan=plan,
            audio_sink=self.router,
            apply_audio=bool(apply_audio),
        )



from __future__ import annotations

from pathlib import Path

import msgspec

from grim.assets import runtime_resources_for
from grim.geom import Vec2
from grim.raylib_api import rl


class SmallFontData(msgspec.Struct, frozen=True):
    widths: list[int]
    texture: rl.Texture
    cell_size: int = 16
    grid: int = 16


SMALL_FONT_FILTER = rl.TextureFilter.TEXTURE_FILTER_POINT


def load_small_font(assets_root: Path) -> SmallFontData:
    return runtime_resources_for(assets_root).small_font


def draw_small_text(font: SmallFontData, text: str, pos: Vec2, color: rl.Color) -> None:
    x_pos = float(int(pos.x))
    y_pos = float(int(pos.y))
    base_x = x_pos
    line_height = float(font.cell_size)
    origin = rl.Vector2(0.0, 0.0)
    for value in text.encode("latin-1", errors="replace"):
        if value == 0x0A:
            x_pos = base_x
            y_pos += line_height
            continue
        if value == 0x0D:
            continue
        width = font.widths[value]
        if width <= 0:
            continue
        col = value % font.grid
        row = value // font.grid
        # Native Grim2D applies a 1/512 UV inset on the DX8 path. Raylib/OpenGL
        # renders visibly cropped glyphs with that bias, so we intentionally use
        # the full glyph rect here.
        src = rl.Rectangle(
            float(col * font.cell_size),
            float(row * font.cell_size),
            float(width),
            float(font.cell_size),
        )
        dst = rl.Rectangle(
            x_pos,
            y_pos,
            float(width),
            float(font.cell_size),
        )
        rl.draw_texture_pro(font.texture, src, dst, origin, 0.0, color)
        x_pos += float(width)


def measure_small_text_height(font: SmallFontData, text: str) -> float:
    line_count = text.count("\n") + 1
    return float(font.cell_size * line_count)


def measure_small_text_width(font: SmallFontData, text: str) -> float:
    """Return the maximum line width for `text` when rendered with `draw_small_text`."""
    x = 0.0
    best = 0.0
    for value in text.encode("latin-1", errors="replace"):
        if value == 0x0A:
            best = max(best, x)
            x = 0.0
            continue
        if value == 0x0D:
            continue
        width = font.widths[value]
        if width <= 0:
            continue
        x += float(width)
    best = max(best, x)
    return best



from __future__ import annotations

from collections.abc import Callable, Sequence
from enum import Enum
from typing import Annotated, Protocol, TypeAlias

import msgspec

from ..local_input import clear_input_edges
from .input import PlayerInput

TypoChar: TypeAlias = Annotated[str, msgspec.Meta(min_length=1, max_length=1)]


class PerkMenuOpenCommand(msgspec.Struct, tag="perk_menu_open", frozen=True):
    player_index: int


class PerkPickCommand(msgspec.Struct, tag="perk_pick", frozen=True):
    player_index: int
    choice_index: int


class TypoCharCommand(msgspec.Struct, tag="typo_char", frozen=True):
    player_index: int
    ch: TypoChar


class TypoBackspaceCommand(msgspec.Struct, tag="typo_backspace", frozen=True):
    player_index: int


class TypoSubmitCommand(msgspec.Struct, tag="typo_submit", frozen=True):
    player_index: int


GameCommand: TypeAlias = (
    PerkMenuOpenCommand
    | PerkPickCommand
    | TypoCharCommand
    | TypoBackspaceCommand
    | TypoSubmitCommand
)


class FrameContext(msgspec.Struct, frozen=True):
    dt_seconds: float
    tick_dt_seconds: float
    frame_index: int
    candidate_ticks: int
    is_networked: bool = False
    is_replay: bool = False


class InputStatus(str, Enum):
    READY = "ready"
    STALLED = "stalled"
    EOS = "eos"


class ResolvedTick(msgspec.Struct, frozen=True):
    tick_index: int
    dt_seconds: float
    inputs: list[PlayerInput] = msgspec.field(default_factory=list)
    commands: list[GameCommand] = msgspec.field(default_factory=list)


class TickSupply(msgspec.Struct, frozen=True):
    status: InputStatus
    tick: ResolvedTick | None = None


class InputProvider(Protocol):
    def begin_frame(self, frame_ctx: FrameContext) -> None: ...

    def pull_tick(self, tick_index: int, default_dt_seconds: float) -> TickSupply: ...

    def supports_command_submission(self) -> bool: ...

    def submit_command(self, command: GameCommand) -> None: ...


LocalInputBuilder: TypeAlias = Callable[[FrameContext], Sequence[PlayerInput]]


class LocalInputProvider:
    """Adapter over local input polling."""

    def __init__(
        self,
        *,
        player_count: int,
        build_inputs: LocalInputBuilder,
    ) -> None:
        self._player_count = max(0, player_count)
        self._build_inputs = build_inputs
        self._pending_commands: list[GameCommand] = []
        self._commands_for_next_tick: list[GameCommand] = []
        self._frame_inputs: list[PlayerInput] = []
        self._edge_inputs: list[PlayerInput] = []
        self._first_tick_pending = False

    def begin_frame(self, frame_ctx: FrameContext) -> None:
        frame_inputs = list(self._build_inputs(frame_ctx))
        self._frame_inputs = list(frame_inputs)
        self._edge_inputs = clear_input_edges(self._frame_inputs)
        self._first_tick_pending = True
        if self._pending_commands:
            self._commands_for_next_tick.extend(self._pending_commands)
            self._pending_commands.clear()

    def pull_tick(self, tick_index: int, default_dt_seconds: float) -> TickSupply:
        tick_index = int(tick_index)
        dt_seconds = float(default_dt_seconds)
        if self._first_tick_pending:
            self._first_tick_pending = False
            inputs = [] if self._player_count <= 0 else list(self._frame_inputs)
            commands = list(self._commands_for_next_tick)
            self._commands_for_next_tick.clear()
            return TickSupply(
                status=InputStatus.READY,
                tick=ResolvedTick(
                    tick_index=tick_index,
                    dt_seconds=dt_seconds,
                    inputs=inputs,
                    commands=commands,
                ),
            )
        inputs = [] if self._player_count <= 0 else list(self._edge_inputs)
        return TickSupply(
            status=InputStatus.READY,
            tick=ResolvedTick(
                tick_index=tick_index,
                dt_seconds=dt_seconds,
                inputs=inputs,
                commands=[],
            ),
        )

    def supports_command_submission(self) -> bool:
        return True

    def submit_command(self, command: GameCommand) -> None:
        self._pending_commands.append(command)



from __future__ import annotations

from .canonical_channels import EntitySamplesSnapshot, RngStreamRow, SimStateSnapshot, TimingSampleRow
from .channel_helpers import ENTITY_SAMPLE_KINDS, EntitySampleRow, entity_rows
from .payloads import BuiltinObject, BuiltinRows, to_builtin_object, to_builtin_value
from .strict_compare import strict_mismatch_payload


def _rng_stream_row_payload(row: RngStreamRow, *, field: str) -> BuiltinObject:
    payload = to_builtin_object(
        {
            "tick_call_index": row.tick_call_index,
            "value_15": row.value_15,
            "state_before_u32": row.state_before_u32,
            "state_after_u32": row.state_after_u32,
            "caller": row.caller,
            "caller_hex": (
                None if row.caller is None else f"0x{row.caller:08x}"
            ),
        },
        field=field,
    )
    return payload


def compare_rng_stream(expected_rows: list[RngStreamRow], actual_rows: list[RngStreamRow]) -> tuple[bool, BuiltinObject | None]:
    exp_keys = [
        (
            row.tick_call_index,
            row.value_15,
            row.state_before_u32,
            row.state_after_u32,
            row.caller,
        )
        for row in expected_rows
    ]
    act_keys = [
        (
            row.tick_call_index,
            row.value_15,
            row.state_before_u32,
            row.state_after_u32,
            row.caller,
        )
        for row in actual_rows
    ]
    max_prefix = min(len(exp_keys), len(act_keys))
    prefix = 0
    while prefix < max_prefix and exp_keys[prefix] == act_keys[prefix]:
        prefix += 1
    if prefix == len(exp_keys) == len(act_keys):
        return True, None
    detail = to_builtin_object(
        {
            "prefix_match_len": prefix,
            "expected_calls": len(exp_keys),
            "actual_calls": len(act_keys),
            "missing_tail": max(0, len(exp_keys) - len(act_keys)),
            "extra_tail": max(0, len(act_keys) - len(exp_keys)),
        },
        field="rng_stream.diff",
    )
    if prefix < len(expected_rows):
        detail["expected_first_mismatch"] = _rng_stream_row_payload(
            expected_rows[prefix],
            field="rng_stream.expected",
        )
    if prefix < len(actual_rows):
        detail["actual_first_mismatch"] = _rng_stream_row_payload(
            actual_rows[prefix],
            field="rng_stream.actual",
        )
    return False, detail


def compare_sim_state(
    expected_obj: SimStateSnapshot,
    actual_obj: SimStateSnapshot,
) -> tuple[bool, BuiltinObject | None]:
    if expected_obj == actual_obj:
        return True, None
    payload, diff_count, pretty = strict_mismatch_payload(expected_obj, actual_obj, root_path="sim_state")
    return False, to_builtin_object(
        {
            "diff_count": int(diff_count),
            "mismatches": payload,
            "pretty": pretty,
        },
        field="sim_state.diff",
    )


def compare_timing_samples(
    expected_rows: list[TimingSampleRow],
    actual_rows: list[TimingSampleRow],
) -> tuple[bool, BuiltinObject | None]:
    if expected_rows == actual_rows:
        return True, None
    payload, diff_count, pretty = strict_mismatch_payload(
        expected_rows,
        actual_rows,
        root_path="timing_samples",
    )
    return False, to_builtin_object(
        {
            "diff_count": int(diff_count),
            "mismatches": payload,
            "pretty": pretty,
        },
        field="timing_samples.diff",
    )


def _rows_by_uid(rows: list[EntitySampleRow]) -> tuple[dict[int, EntitySampleRow], int, list[int]]:
    by_uid: dict[int, EntitySampleRow] = {}
    duplicate_uids: list[int] = []
    for row in rows:
        uid = int(row.uid)
        if uid in by_uid:
            duplicate_uids.append(uid)
        by_uid[uid] = row
    return by_uid, len(rows), duplicate_uids


def compare_entity_samples(
    expected_obj: EntitySamplesSnapshot,
    actual_obj: EntitySamplesSnapshot,
) -> tuple[bool, BuiltinObject | None]:
    if expected_obj == actual_obj:
        return True, None

    detail: BuiltinObject = {}
    row_diffs: BuiltinRows = []
    for kind in ENTITY_SAMPLE_KINDS:
        exp_map, exp_total, exp_dupes = _rows_by_uid(entity_rows(expected_obj, kind=kind))
        act_map, act_total, act_dupes = _rows_by_uid(entity_rows(actual_obj, kind=kind))
        if exp_total != act_total:
            detail[f"{kind}_count"] = {"expected": exp_total, "actual": act_total}
        if exp_dupes or act_dupes:
            detail[f"{kind}_duplicate_uids"] = {
                "expected": exp_dupes,
                "actual": act_dupes,
            }

        missing = sorted(uid for uid in exp_map if uid not in act_map)
        extra = sorted(uid for uid in act_map if uid not in exp_map)
        if missing or extra:
            detail[f"{kind}_uids"] = {
                "missing": missing,
                "extra": extra,
            }

        for uid in sorted(set(exp_map) & set(act_map)):
            expected_row = exp_map[uid]
            actual_row = act_map[uid]
            if expected_row == actual_row:
                continue
            row_path = f"entity_samples.{kind}[uid={uid}]"
            payload, diff_count, pretty = strict_mismatch_payload(
                expected_row,
                actual_row,
                root_path=row_path,
            )
            row_diffs.append(
                {
                    "path": row_path,
                    "diff_count": int(diff_count),
                    "mismatches": payload,
                    "pretty": pretty,
                },
            )

    if row_diffs:
        detail["row_diffs"] = to_builtin_value(row_diffs, field="entity_samples.row_diffs")
    return (len(detail) == 0), (None if len(detail) == 0 else detail)



from __future__ import annotations

from collections.abc import Callable

from ..terrain_slots import TerrainSlotTriplet, terrain_slots_for_quest
from ..weapons import WeaponId
from .level import QuestLevel
from .types import QuestBuilder, QuestDefinition

_QUESTS: dict[QuestLevel, QuestDefinition] = {}


def register_quest(
    *,
    level: str,
    title: str,
    time_limit_ms: int,
    start_weapon_id: WeaponId,
    unlock_perk_id: int | None = None,
    unlock_weapon_id: WeaponId | None = None,
    terrain_slots: TerrainSlotTriplet | None = None,
) -> Callable[[QuestBuilder], QuestBuilder]:
    def _builder_name(builder_fn: QuestBuilder) -> str:
        return str(getattr(builder_fn, "__name__", repr(builder_fn)))

    def decorator(builder: QuestBuilder) -> QuestBuilder:
        quest_level = QuestLevel.parse(level)
        resolved_terrain_slots = (
            terrain_slots if terrain_slots is not None else terrain_slots_for_quest(quest_level)
        )
        normalized_unlock_weapon_id = WeaponId(unlock_weapon_id) if unlock_weapon_id is not None else None
        quest = QuestDefinition(
            level=quest_level,
            title=title,
            builder=builder,
            time_limit_ms=time_limit_ms,
            start_weapon_id=WeaponId(start_weapon_id),
            unlock_perk_id=unlock_perk_id,
            unlock_weapon_id=normalized_unlock_weapon_id,
            terrain_slots=resolved_terrain_slots,
        )
        existing = _QUESTS.get(quest.level)
        if existing is not None:
            raise ValueError(
                f"duplicate quest level {quest.level.text}: {_builder_name(existing.builder)} vs {_builder_name(builder)}",
            )
        _QUESTS[quest.level] = quest
        return builder

    return decorator


def all_quests() -> list[QuestDefinition]:
    return sorted(_QUESTS.values(), key=lambda quest: quest.level.global_index)


def quest_by_level(level: QuestLevel) -> QuestDefinition | None:
    return _QUESTS.get(level)



from __future__ import annotations

import json
from collections.abc import Mapping, Sequence
from typing import cast

import msgspec

from .payloads import BuiltinRows, BuiltinValue, to_builtin_rows, to_builtin_value


class FieldMismatch(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    path: str
    kind: str
    expected: BuiltinValue | None = None
    actual: BuiltinValue | None = None


def _path_or_root(path: str) -> str:
    return path if path else "<root>"


def _path_attr(path: str, name: str) -> str:
    return str(name) if not path else f"{path}.{name}"


def _path_index(path: str, index: int) -> str:
    return f"{path}[{int(index)}]" if path else f"[{int(index)}]"


def _path_key(path: str, key: object) -> str:
    if isinstance(key, str) and key.isidentifier():
        return _path_attr(path, key)
    key_text = json.dumps(key, sort_keys=True, default=repr)
    return f"{path}[{key_text}]" if path else f"[{key_text}]"


def _is_struct_instance(value: object) -> bool:
    fields = getattr(type(value), "__struct_fields__", None)
    return isinstance(fields, tuple)


def _struct_field_names(value: object) -> tuple[str, ...]:
    fields = getattr(type(value), "__struct_fields__", None)
    if isinstance(fields, tuple) and all(isinstance(name, str) for name in fields):
        return tuple(str(name) for name in fields)
    return ()


def _is_sequence(value: object) -> bool:
    return isinstance(value, Sequence) and not isinstance(value, (str, bytes, bytearray, memoryview))


def _append(
    out: list[FieldMismatch],
    *,
    path: str,
    kind: str,
    expected: object | None = None,
    actual: object | None = None,
) -> None:
    out.append(
        FieldMismatch(
            path=_path_or_root(path),
            kind=str(kind),
            expected=to_builtin_value(expected, field=f"{_path_or_root(path)}.expected") if expected is not None else None,
            actual=to_builtin_value(actual, field=f"{_path_or_root(path)}.actual") if actual is not None else None,
        ),
    )


def _collect(expected: object, actual: object, *, path: str, out: list[FieldMismatch]) -> None:
    if type(expected) is type(actual) and expected == actual:
        return

    if type(expected) is not type(actual):
        _append(out, path=path, kind="type_mismatch", expected=type(expected).__name__, actual=type(actual).__name__)
        return

    if _is_struct_instance(expected):
        for name in _struct_field_names(expected):
            _collect(
                getattr(expected, name),
                getattr(actual, name),
                path=_path_attr(path, name),
                out=out,
            )
        return

    if isinstance(expected, Mapping):
        expected_map = cast("Mapping[object, object]", expected)
        actual_map = cast("Mapping[object, object]", actual)
        keys = sorted(set(expected_map.keys()) | set(actual_map.keys()), key=lambda key: repr(key))
        for key in keys:
            key_path = _path_key(path, key)
            if key not in expected_map:
                _append(out, path=key_path, kind="missing_expected", actual=actual_map[key])
                continue
            if key not in actual_map:
                _append(out, path=key_path, kind="missing_actual", expected=expected_map[key])
                continue
            _collect(expected_map[key], actual_map[key], path=key_path, out=out)
        return

    if _is_sequence(expected):
        expected_seq = cast("Sequence[object]", expected)
        actual_seq = cast("Sequence[object]", actual)
        if len(expected_seq) != len(actual_seq):
            _append(out, path=path, kind="length_mismatch", expected=len(expected_seq), actual=len(actual_seq))
        shared_len = min(len(expected_seq), len(actual_seq))
        for index in range(shared_len):
            _collect(expected_seq[index], actual_seq[index], path=_path_index(path, index), out=out)
        for index in range(shared_len, len(expected_seq)):
            _append(out, path=_path_index(path, index), kind="missing_actual", expected=expected_seq[index])
        for index in range(shared_len, len(actual_seq)):
            _append(out, path=_path_index(path, index), kind="missing_expected", actual=actual_seq[index])
        return

    _append(out, path=path, kind="value_mismatch", expected=expected, actual=actual)


def strict_mismatches(expected: object, actual: object, *, root_path: str = "") -> list[FieldMismatch]:
    out: list[FieldMismatch] = []
    _collect(expected, actual, path=str(root_path), out=out)
    return out


def strict_mismatch_payload(expected: object, actual: object, *, root_path: str = "") -> tuple[BuiltinRows, int, str]:
    mismatches = strict_mismatches(expected, actual, root_path=root_path)
    payload = to_builtin_rows(mismatches, field=f"{root_path or 'mismatches'}")
    return payload, len(payload), json.dumps(payload, sort_keys=True, indent=2, default=repr)



from __future__ import annotations

from ..replay.checkpoints import ReplayCheckpoint
from .canonical_channels import (
    BonusEntitySample,
    CreatureEntitySample,
    EntitySamplesSnapshot,
    ProjectileEntitySample,
    RngStreamRow,
    SecondaryProjectileEntitySample,
    SimStateSnapshot,
    TimingSampleRow,
)
from .schema import TickRecord
from .trace import TraceError

ENTITY_SAMPLE_KINDS = ("creatures", "projectiles", "secondary_projectiles", "bonuses")

EntitySampleRow = CreatureEntitySample | ProjectileEntitySample | SecondaryProjectileEntitySample | BonusEntitySample


def _require_row(row: TickRecord | None, *, channel_name: str) -> TickRecord:
    if row is None:
        raise TraceError(f"missing tick row for {channel_name} channel")
    return row


def checkpoint_channel(row: TickRecord | None) -> ReplayCheckpoint | None:
    return None if row is None else row.channels.checkpoint


def checkpoint_channel_required(row: TickRecord | None) -> ReplayCheckpoint:
    return _require_row(row, channel_name="checkpoint").channels.checkpoint


def sim_state_channel(row: TickRecord | None) -> SimStateSnapshot | None:
    return None if row is None else row.channels.sim_state


def sim_state_channel_required(row: TickRecord | None) -> SimStateSnapshot:
    return _require_row(row, channel_name="sim_state").channels.sim_state


def entity_samples_channel(row: TickRecord | None) -> EntitySamplesSnapshot | None:
    return None if row is None else row.channels.entity_samples


def entity_samples_channel_required(row: TickRecord | None) -> EntitySamplesSnapshot:
    return _require_row(row, channel_name="entity_samples").channels.entity_samples


def rng_stream_channel_required(row: TickRecord | None) -> list[RngStreamRow]:
    return list(_require_row(row, channel_name="rng_stream").channels.rng_stream)

def timing_samples_channel_required(row: TickRecord | None) -> list[TimingSampleRow]:
    return list(_require_row(row, channel_name="timing_samples").channels.timing_samples)


def entity_rows(samples: EntitySamplesSnapshot, *, kind: str) -> list[EntitySampleRow]:
    match str(kind):
        case "creatures":
            return [*samples.creatures]
        case "projectiles":
            return [*samples.projectiles]
        case "secondary_projectiles":
            return [*samples.secondary_projectiles]
        case "bonuses":
            return [*samples.bonuses]
        case _:
            raise ValueError(f"unsupported entity sample kind: {kind!r}")



from __future__ import annotations

import math
from collections.abc import Callable, MutableSequence, Sequence
from typing import TYPE_CHECKING

import msgspec

from grim.geom import Vec2
from grim.rand import CrandLike
from grim.sfx_map import SfxId

from ...creatures.damage_types import CreatureDamageType
from ...creatures.lifecycle import creature_lifecycle_is_collidable
from ...creatures.spawn import CreatureFlags
from ...effects import EffectPool
from ...math_parity import f32
from ...owner_ref import OwnerRef
from ...rng_caller_static import RngCallerStatic
from ...weapons import weapon_entry_for_projectile_type_id
from ..effects import (
    _spawn_ion_hit_effects,
    _spawn_plasma_cannon_hit_effects,
    _spawn_shrinkifier_hit_effects,
    _spawn_splitter_hit_effects,
)
from ..types import (
    Projectile,
    ProjectileTemplateId,
)
from .collision import _apply_damage_to_creature, _hit_radius_for

if TYPE_CHECKING:
    from ...creatures.runtime import CreatureState
    from ...gameplay import GameplayState
    from .projectile_pool import ProjectilePool


class _ProjectileUpdateCtx(msgspec.Struct):
    pool: ProjectilePool
    creatures: Sequence[CreatureState]
    dt: float
    ion_scale: float
    detail_preset: int
    rng: CrandLike
    runtime_state: GameplayState | None
    effects: EffectPool | None
    sfx_queue: MutableSequence[SfxId] | None


class _ProjectileHitInfo(msgspec.Struct):
    proj_index: int
    proj: Projectile
    hit_idx: int
    move: Vec2
    target: Vec2


class _ProjectileHitPerkCtx(msgspec.Struct):
    proj: Projectile
    creature: CreatureState
    rng: CrandLike
    owner_perk_active: Callable[[OwnerRef, int], bool]
    poison_idx: int


_ProjectileHitPerkHook = Callable[[_ProjectileHitPerkCtx], None]


def _projectile_hit_perk_poison_bullets(ctx: _ProjectileHitPerkCtx) -> None:
    if (
        ctx.owner_perk_active(ctx.proj.owner, int(ctx.poison_idx))
        and (ctx.rng.rand(caller=RngCallerStatic.PROJECTILE_UPDATE_POISON_BULLETS_GATE) & 7) == 1
    ):
        ctx.creature.flags |= CreatureFlags.SELF_DAMAGE_TICK


_PROJECTILE_HIT_PERK_HOOKS: tuple[_ProjectileHitPerkHook, ...] = (_projectile_hit_perk_poison_bullets,)


def _life_timer_sub_f32(life_timer: float, amount: float) -> float:
    return float(f32(float(life_timer) - float(amount)))


def _linger_default(ctx: _ProjectileUpdateCtx, proj: Projectile) -> None:
    proj.life_timer = _life_timer_sub_f32(float(proj.life_timer), float(ctx.dt))


def _linger_gauss_gun(ctx: _ProjectileUpdateCtx, proj: Projectile) -> None:
    proj.life_timer = _life_timer_sub_f32(float(proj.life_timer), float(ctx.dt) * 0.1)


def _linger_ion_aoe(
    ctx: _ProjectileUpdateCtx,
    proj: Projectile,
    *,
    life_decay_scale: float,
    damage_per_second: float,
    base_radius: float,
) -> None:
    proj.life_timer = _life_timer_sub_f32(float(proj.life_timer), float(ctx.dt) * float(life_decay_scale))
    damage = float(ctx.dt) * float(damage_per_second)
    radius = float(ctx.ion_scale) * float(base_radius)
    for creature_idx, creature in enumerate(ctx.creatures):
        if not creature.active:
            continue
        if not creature_lifecycle_is_collidable(creature.lifecycle_stage):
            continue
        creature_radius = _hit_radius_for(creature)
        hit_r = radius + creature_radius
        if Vec2.distance_sq(proj.pos, creature.pos) <= hit_r * hit_r:
            _apply_damage_to_creature(
                ctx.creatures,
                creature_idx,
                damage,
                damage_type=CreatureDamageType.ION,
                impulse=Vec2(),
                owner=proj.owner,
                apply_creature_damage=ctx.pool.creature_damage_applier,
            )


def _linger_ion_minigun(ctx: _ProjectileUpdateCtx, proj: Projectile) -> None:
    _linger_ion_aoe(
        ctx,
        proj,
        life_decay_scale=1.0,
        damage_per_second=40.0,
        base_radius=60.0,
    )


def _linger_ion_rifle(ctx: _ProjectileUpdateCtx, proj: Projectile) -> None:
    _linger_ion_aoe(
        ctx,
        proj,
        life_decay_scale=1.0,
        damage_per_second=100.0,
        base_radius=88.0,
    )


def _linger_ion_cannon(ctx: _ProjectileUpdateCtx, proj: Projectile) -> None:
    _linger_ion_aoe(
        ctx,
        proj,
        life_decay_scale=0.7,
        damage_per_second=300.0,
        base_radius=128.0,
    )


def _pre_hit_splitter(ctx: _ProjectileUpdateCtx, proj: Projectile, hit_idx: int) -> None:
    _spawn_splitter_hit_effects(
        ctx.effects,
        pos=proj.pos,
        rng=ctx.rng,
        detail_preset=ctx.detail_preset,
    )
    # Native player-hit checks key off non-player ownership; creature-owned splitters
    # always satisfy this, so they can hit players even when the parent was local-owned.
    split_hits_players = True
    ctx.pool.spawn(
        pos=proj.pos,
        angle=proj.angle - 1.0471976,
        type_id=ProjectileTemplateId.SPLITTER_GUN,
        owner=OwnerRef.from_creature(int(hit_idx)),
        travel_budget=proj.travel_budget,
        hits_players=split_hits_players,
    )
    ctx.pool.spawn(
        pos=proj.pos,
        angle=proj.angle + 1.0471976,
        type_id=ProjectileTemplateId.SPLITTER_GUN,
        owner=OwnerRef.from_creature(int(hit_idx)),
        travel_budget=proj.travel_budget,
        hits_players=split_hits_players,
    )


def _post_hit_ion_common(ctx: _ProjectileUpdateCtx, hit: _ProjectileHitInfo) -> None:
    _spawn_ion_hit_effects(
        ctx.effects,
        ctx.sfx_queue,
        type_id=ProjectileTemplateId(hit.proj.type_id),
        pos=hit.proj.pos,
        rng=ctx.rng,
        detail_preset=ctx.detail_preset,
    )


def _post_hit_ion_rifle(ctx: _ProjectileUpdateCtx, hit: _ProjectileHitInfo) -> None:
    runtime_state = ctx.runtime_state
    creatures = ctx.creatures
    hit_creature = int(hit.hit_idx)
    if (
        runtime_state is not None
        and runtime_state.shock_chain_projectile_id == hit.proj_index
        and 0 <= hit_creature < len(creatures)
    ):
        links_left = int(runtime_state.shock_chain_links_left)
        if links_left > 0 and creatures:
            runtime_state.shock_chain_links_left = links_left - 1

            origin_pos = hit.proj.pos
            min_dist_sq = 100.0 * 100.0

            best_idx = 0 if bool(runtime_state.preserve_bugs) else -1
            best_dist_sq = 1e12
            for creature_id, creature in enumerate(creatures):
                if creature_id == hit_creature:
                    continue
                if not creature.active:
                    continue
                d_sq = Vec2.distance_sq(origin_pos, creature.pos)
                if d_sq <= min_dist_sq:
                    continue
                if d_sq < best_dist_sq:
                    best_dist_sq = d_sq
                    best_idx = creature_id

            if best_idx < 0:
                _post_hit_ion_common(ctx, hit)
                return

            origin = creatures[hit_creature]
            target = creatures[best_idx]
            angle = (target.pos - origin.pos).to_heading()

            prev_guard = bool(runtime_state.bonus_spawn_guard)
            runtime_state.bonus_spawn_guard = True
            try:
                proj_id = ctx.pool.spawn(
                    pos=origin_pos,
                    angle=angle,
                    type_id=ProjectileTemplateId(hit.proj.type_id),
                    owner=OwnerRef.from_creature(hit_creature),
                    travel_budget=hit.proj.travel_budget,
                )
            finally:
                runtime_state.bonus_spawn_guard = prev_guard
            runtime_state.shock_chain_projectile_id = proj_id
    _post_hit_ion_common(ctx, hit)


def _post_hit_plasma_cannon(ctx: _ProjectileUpdateCtx, hit: _ProjectileHitInfo) -> None:
    creature = ctx.creatures[int(hit.hit_idx)]
    size = float(creature.size)
    ring_radius = size * 0.5 + 1.0

    plasma_entry = weapon_entry_for_projectile_type_id(ProjectileTemplateId.PLASMA_RIFLE)
    plasma_meta = float(plasma_entry.travel_budget)

    runtime_state = ctx.runtime_state
    prev_guard = False
    if runtime_state is not None:
        prev_guard = bool(runtime_state.bonus_spawn_guard)
        runtime_state.bonus_spawn_guard = True
    try:
        for ring_idx in range(12):
            ring_angle = float(ring_idx) * (math.pi / 6.0)
            ring_offset = Vec2.from_angle(ring_angle) * ring_radius
            ctx.pool.spawn(
                pos=hit.proj.pos + ring_offset,
                angle=ring_angle,
                type_id=ProjectileTemplateId.PLASMA_RIFLE,
                owner=OwnerRef.from_local_player(0),
                travel_budget=plasma_meta,
            )
    finally:
        if runtime_state is not None:
            runtime_state.bonus_spawn_guard = prev_guard

    _spawn_plasma_cannon_hit_effects(
        ctx.effects,
        ctx.sfx_queue,
        pos=hit.proj.pos,
        detail_preset=ctx.detail_preset,
    )


def _post_hit_shrinkifier(ctx: _ProjectileUpdateCtx, hit: _ProjectileHitInfo) -> None:
    _spawn_shrinkifier_hit_effects(
        ctx.effects,
        pos=hit.proj.pos,
        rng=ctx.rng,
        detail_preset=ctx.detail_preset,
    )

    creature = ctx.creatures[int(hit.hit_idx)]
    new_size = float(creature.size) * 0.65
    creature.size = new_size
    if new_size < 16.0:
        _apply_damage_to_creature(
            ctx.creatures,
            int(hit.hit_idx),
            float(creature.hp) + 1.0,
            damage_type=CreatureDamageType.BULLET,
            impulse=Vec2(),
            owner=hit.proj.owner,
            apply_creature_damage=ctx.pool.creature_damage_applier,
        )
    hit.proj.life_timer = 0.25


def _post_hit_pulse_gun(ctx: _ProjectileUpdateCtx, hit: _ProjectileHitInfo) -> None:
    creature = ctx.creatures[int(hit.hit_idx)]
    creature.pos = creature.pos + hit.move * 3.0


def _post_hit_plague_spreader(ctx: _ProjectileUpdateCtx, hit: _ProjectileHitInfo) -> None:
    creature = ctx.creatures[int(hit.hit_idx)]
    creature.plague_infected = True



from __future__ import annotations

from .types import GameConfig

__all__ = ["GameConfig", "run_game"]


def run_game(config: GameConfig) -> None:
    # Keep package import side effects minimal so UI modules can import
    # `crimson.game.types` without pulling in the full runtime graph.
    from .runtime import run_game as _run_game

    _run_game(config)



from __future__ import annotations

from dataclasses import dataclass
from typing import Protocol

from ...sim.batch_apply import PresentationTickOutput, SimMetadataSink, apply_tick_to_sim
from ...sim.clock import FixedStepClock
from ...sim.frame_pump import advance_tick_runner_frame
from ...sim.hooks import TickResult
from ...sim.input_providers import FrameContext, InputStatus
from ...sim.tick_runner import TickBatchResult


class PlaybackFrameDriver(Protocol):
    def step_tick(self, tick_index: int) -> TickResult: ...


@dataclass(frozen=True, slots=True)
class PlaybackFrameAdvance:
    outputs: tuple[PresentationTickOutput, ...]
    tick_results: tuple[TickResult, ...]
    frame_index: int
    next_tick_index: int
    ticks_requested: int
    ticks_completed: int


class _PlaybackStepRunner:
    def __init__(self, *, driver: PlaybackFrameDriver, tick_limit: int) -> None:
        self._driver = driver
        self._tick_limit = int(tick_limit)

    def begin_frame(self, frame_ctx: FrameContext) -> None:
        _ = frame_ctx

    def advance_ticks(
        self,
        *,
        start_tick: int,
        ticks_requested: int,
        tick_dt: float,
    ) -> TickBatchResult:
        _ = tick_dt
        completed_results: list[TickResult] = []
        next_tick_index = int(start_tick)
        batch_status = InputStatus.READY
        while len(completed_results) < max(0, int(ticks_requested)):
            if next_tick_index >= int(self._tick_limit):
                batch_status = InputStatus.EOS
                break
            completed_results.append(self._driver.step_tick(next_tick_index))
            next_tick_index += 1

        return TickBatchResult(
            ticks_completed=len(completed_results),
            batch_status=batch_status,
            next_tick_index=int(next_tick_index),
            completed_results=completed_results,
        )


def advance_playback_frame(
    *,
    driver: PlaybackFrameDriver,
    sim_world: SimMetadataSink,
    clock: FixedStepClock,
    start_tick: int,
    frame_index: int,
    dt_seconds: float,
    max_ticks: int | None,
    tick_limit: int,
    game_tune_started: bool,
) -> PlaybackFrameAdvance:
    ticks_requested = int(clock.advance(float(dt_seconds)))
    if max_ticks is not None:
        ticks_requested = min(int(ticks_requested), max(0, int(max_ticks)))

    advance = advance_tick_runner_frame(
        runner=_PlaybackStepRunner(driver=driver, tick_limit=int(tick_limit)),
        start_tick=int(start_tick),
        frame_index=int(frame_index),
        ticks_requested=int(ticks_requested),
        dt_seconds=float(dt_seconds),
        tick_dt_seconds=float(clock.dt_tick),
        is_networked=False,
        is_replay=True,
        refund_clock=clock,
    )
    tick_results = list(advance.batch.completed_results)
    outputs: list[PresentationTickOutput] = []
    for tick_result in tick_results:
        apply_tick_to_sim(
            sim_world=sim_world,
            step=tick_result.payload.step,
            game_tune_started=bool(game_tune_started),
        )
        outputs.append(PresentationTickOutput(
            tick_index=int(tick_result.source_tick.tick_index),
            dt_sim=float(tick_result.payload.step.dt_sim),
            presentation=tick_result.payload.step.presentation,
            terrain_fx=tick_result.payload.step.terrain_fx,
        ))

    return PlaybackFrameAdvance(
        outputs=tuple(outputs),
        tick_results=tuple(tick_results),
        frame_index=int(advance.frame_index),
        next_tick_index=int(advance.next_tick_index),
        ticks_requested=int(advance.ticks_requested),
        ticks_completed=int(advance.batch.ticks_completed),
    )



from __future__ import annotations

import math
from collections.abc import MutableSequence, Sequence
from typing import TYPE_CHECKING

import msgspec

from grim.color import RGBA
from grim.geom import Vec2
from grim.rand import Crand
from grim.sfx_map import SfxId

from ...creatures.damage_types import CreatureDamageType
from ...creatures.lifecycle import creature_lifecycle_is_alive, creature_lifecycle_is_collidable
from ...effects import EffectPool, FxQueue, SpriteEffectPool
from ...effects_atlas import EffectId
from ...math_parity import f32
from ...owner_ref import OwnerRef
from ...rng_caller_static import RngCallerStatic
from ..types import (
    SECONDARY_PROJECTILE_POOL_SIZE,
    CreatureDamageApplier,
    SecondaryDetonationKillHandler,
    SecondaryProjectile,
    SecondaryProjectileTypeId,
)
from .collision import _apply_damage_to_creature, _creature_find_nearest_for_secondary, _within_native_find_radius
from .secondary_rules import (
    DetonationRule,
    HomingRocketRule,
    RocketMinigunRule,
    RocketRule,
    secondary_rule_for_type_id,
)
from .spatial_hash import CreatureSpatialHash

if TYPE_CHECKING:
    from ...creatures.runtime import CreatureState
    from ...gameplay import GameplayState


_SECONDARY_PRE_HIT_DECAL_CALLERS = (
    (
        RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DX_1,
        RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DY_1,
    ),
    (
        RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DX_2,
        RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DY_2,
    ),
    (
        RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DX_3,
        RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DY_3,
    ),
)


class SecondarySpawnSpec(msgspec.Struct, frozen=True):
    pos: Vec2
    angle: float
    type_id: SecondaryProjectileTypeId
    owner: OwnerRef = msgspec.field(default_factory=lambda: OwnerRef.from_local_player(0))
    time_to_live: float = 2.0
    target_hint: Vec2 | None = None
    creatures: Sequence[CreatureState] | None = None
    preserve_bugs: bool = False


class SecondaryStepCtx(msgspec.Struct, frozen=True):
    dt: float
    creatures: Sequence[CreatureState]
    runtime_state: GameplayState | None = None
    fx_queue: FxQueue | None = None
    detail_preset: int = 5
    on_detonation_kill: SecondaryDetonationKillHandler | None = None


class SecondaryProjectilePool:
    def __init__(self, *, size: int = SECONDARY_PROJECTILE_POOL_SIZE) -> None:
        self._entries = [SecondaryProjectile() for _ in range(size)]
        self._creature_damage_applier: CreatureDamageApplier | None = None

    @property
    def entries(self) -> list[SecondaryProjectile]:
        return self._entries

    @property
    def creature_damage_applier(self) -> CreatureDamageApplier | None:
        return self._creature_damage_applier

    @creature_damage_applier.setter
    def creature_damage_applier(self, value: CreatureDamageApplier | None) -> None:
        self._creature_damage_applier = value

    def reset(self) -> None:
        for entry in self._entries:
            entry.active = False

    def spawn_from_spec(self, spec: SecondarySpawnSpec) -> int:
        pos = spec.pos
        angle = float(spec.angle)
        type_id = SecondaryProjectileTypeId(spec.type_id)
        owner = spec.owner
        time_to_live = float(spec.time_to_live)
        target_hint = spec.target_hint
        creatures = spec.creatures
        preserve_bugs = bool(spec.preserve_bugs)

        index = None
        for i, entry in enumerate(self._entries):
            if not entry.active:
                index = i
                break
        if index is None:
            index = len(self._entries) - 1

        entry = self._entries[index]
        entry.active = True
        entry.angle = float(angle)
        entry.type_id = type_id
        entry.pos = pos
        entry.owner = owner
        entry.target_id = -1
        entry.trail_timer = 0.0
        entry.vel = Vec2()
        entry.detonation_t = 0.0
        entry.detonation_scale = 1.0

        rule = secondary_rule_for_type_id(type_id)
        match rule:
            case DetonationRule():
                # Detonation uses explicit timer/scale fields now.
                entry.detonation_t = 0.0
                entry.detonation_scale = float(time_to_live)
                entry.speed = float(f32(float(time_to_live)))
                return index
            case (
                RocketRule(base_speed=base_speed)
                | HomingRocketRule(base_speed=base_speed)
                | RocketMinigunRule(
                    base_speed=base_speed,
                )
            ):
                entry.vel = Vec2.from_heading(float(angle)) * float(base_speed)
                entry.speed = float(f32(float(time_to_live)))

        if isinstance(rule, HomingRocketRule):
            # Native `fx_spawn_secondary_projectile` seeds seeker target_id at spawn via
            # `creature_find_nearest(&player_aim_x, -1, 0.0)`.
            if creatures is not None:
                origin = target_hint if target_hint is not None else pos
                entry.target_id = _creature_find_nearest_for_secondary(
                    creatures=creatures,
                    origin=origin,
                    preserve_bugs=preserve_bugs,
                )

        return index

    def iter_active(self) -> list[SecondaryProjectile]:
        return [entry for entry in self._entries if entry.active]

    def step(self, ctx: SecondaryStepCtx) -> None:
        """Update the secondary projectile pool subset (types 1/2/4 + detonation type 3)."""
        dt = float(ctx.dt)
        creatures = ctx.creatures
        runtime_state = ctx.runtime_state
        fx_queue = ctx.fx_queue
        detail_preset = int(ctx.detail_preset)
        on_detonation_kill = ctx.on_detonation_kill

        if dt <= 0.0:
            return

        def _apply_secondary_damage(
            creature_index: int,
            damage: float,
            *,
            owner: OwnerRef,
            impulse: Vec2 = Vec2(),
        ) -> None:
            _apply_damage_to_creature(
                creatures,
                int(creature_index),
                float(damage),
                damage_type=CreatureDamageType.EXPLOSION,
                impulse=impulse,
                owner=owner,
                apply_creature_damage=self._creature_damage_applier,
            )

        rng = Crand(0)
        freeze_active = False
        effects: EffectPool | None = None
        sprite_effects: SpriteEffectPool | None = None
        sfx_queue: MutableSequence[SfxId] | None = None
        if runtime_state is not None:
            rng = runtime_state.rng
            freeze_active = float(runtime_state.bonuses.freeze) > 0.0
            effects = runtime_state.effects
            sprite_effects = runtime_state.sprite_effects
            sfx_queue = runtime_state.sfx_queue

        def _creature_is_collidable(creature: CreatureState) -> bool:
            if not creature.active:
                return False
            return creature_lifecycle_is_collidable(creature.lifecycle_stage)

        creature_spatial = CreatureSpatialHash(creatures=creatures, is_collidable=_creature_is_collidable)

        for entry in self._entries:
            if not entry.active:
                continue

            rule = secondary_rule_for_type_id(SecondaryProjectileTypeId(entry.type_id))

            if isinstance(rule, DetonationRule):
                if runtime_state is not None:
                    runtime_state.camera_shake_pulses = 4

                entry.detonation_t += dt * 3.0
                t = float(entry.detonation_t)
                scale = float(entry.detonation_scale)
                if t > 1.0:
                    if fx_queue is not None:
                        fx_queue.add(
                            effect_id=int(EffectId.AURA),
                            pos=entry.pos,
                            width=float(scale) * 256.0,
                            height=float(scale) * 256.0,
                            rotation=0.0,
                            rgba=RGBA(0.0, 0.0, 0.0, 0.25),
                        )
                    entry.active = False

                radius = scale * t * 80.0
                radius_sq = radius * radius
                damage = dt * scale * 700.0
                for creature_idx in creature_spatial.candidate_indices(pos=entry.pos, radius=float(radius)):
                    creature = creatures[int(creature_idx)]
                    if not _creature_is_collidable(creature):
                        continue
                    if creature.hp <= 0.0:
                        continue
                    d_sq = Vec2.distance_sq(entry.pos, creature.pos)
                    if d_sq < radius_sq:
                        hp_before = float(creature.hp)
                        impulse_dir = entry.pos.direction_to(creature.pos)
                        impulse = impulse_dir * 0.1
                        _apply_secondary_damage(
                            creature_idx,
                            damage,
                            owner=entry.owner,
                            impulse=impulse,
                        )
                        creature_spatial.sync_index(int(creature_idx))
                        if on_detonation_kill is not None and hp_before > 0.0 and float(creature.hp) <= 0.0:
                            # Native detonation AoE does an extra two random decals and a
                            # second `creature_handle_death` call after the killing hit.
                            if fx_queue is not None:
                                fx_queue.add_random(pos=creature.pos, rng=rng)
                                fx_queue.add_random(pos=creature.pos, rng=rng)
                            on_detonation_kill(int(creature_idx))
                continue

            if not isinstance(rule, (RocketRule, HomingRocketRule, RocketMinigunRule)):
                continue

            # Move.
            entry.pos = entry.pos + entry.vel * dt

            # Update velocity + countdown.
            speed_mag = entry.vel.length()
            match rule:
                case RocketRule(
                    accel_factor_scale=accel_factor_scale, speed_cap=speed_cap, ttl_decay_scale=ttl_decay_scale,
                ):
                    if speed_mag < float(speed_cap):
                        factor = 1.0 + dt * float(accel_factor_scale)
                        entry.vel = entry.vel * factor
                    entry.speed = float(f32(float(entry.speed) - float(dt) * float(ttl_decay_scale)))
                case RocketMinigunRule(
                    accel_factor_scale=accel_factor_scale,
                    speed_cap=speed_cap,
                    ttl_decay_scale=ttl_decay_scale,
                ):
                    if speed_mag < float(speed_cap):
                        factor = 1.0 + dt * float(accel_factor_scale)
                        entry.vel = entry.vel * factor
                    entry.speed = float(f32(float(entry.speed) - float(dt) * float(ttl_decay_scale)))
                case HomingRocketRule(
                    target_accel=target_accel, max_velocity=max_velocity, ttl_decay_scale=ttl_decay_scale,
                ):
                    # Type 2: homing projectile.
                    target_id = entry.target_id
                    if not (0 <= target_id < len(creatures)) or not creatures[target_id].active:
                        entry.target_id = _creature_find_nearest_for_secondary(
                            creatures=creatures,
                            origin=entry.pos,
                            preserve_bugs=bool(runtime_state.preserve_bugs) if runtime_state is not None else False,
                        )
                        target_id = entry.target_id

                    if 0 <= target_id < len(creatures):
                        target = creatures[target_id]
                        to_target = target.pos - entry.pos
                        target_dir, dist = to_target.normalized_with_length()
                        if dist > 1e-6:
                            entry.angle = to_target.to_heading()
                            accel = target_dir * (dt * float(target_accel))
                            next_velocity = entry.vel + accel
                            if next_velocity.length() <= float(max_velocity):
                                entry.vel = next_velocity

                    entry.speed = float(f32(float(entry.speed) - float(dt) * float(ttl_decay_scale)))

            # Rocket smoke trail (`trail_timer` in crimsonland.exe).
            trail_decay = float(f32((abs(entry.vel.x) + abs(entry.vel.y)) * dt * 0.01))
            entry.trail_timer = float(f32(float(entry.trail_timer) - float(trail_decay)))
            if float(entry.trail_timer) < 0.0:
                direction = Vec2.from_heading(entry.angle)
                spawn_pos = entry.pos - direction * 9.0
                trail_velocity = Vec2.from_heading(entry.angle + math.pi) * 90.0
                if sprite_effects is not None:
                    sprite_effects.spawn(
                        pos=spawn_pos,
                        vel=trail_velocity,
                        scale=14.0,
                        color=RGBA(1.0, 1.0, 1.0, 0.25),
                    )
                entry.trail_timer = float(f32(0.06))

            # projectile_update uses creature_find_in_radius(..., 8.0, ...)
            hit_idx: int | None = None
            for idx in creature_spatial.candidate_indices(pos=entry.pos, radius=8.0):
                creature = creatures[int(idx)]
                if not _creature_is_collidable(creature):
                    continue
                if _within_native_find_radius(
                    origin=entry.pos,
                    target=creature.pos,
                    radius=8.0,
                    target_size=float(creature.size),
                ):
                    hit_idx = idx
                    break
            if hit_idx is not None:
                if runtime_state is not None:
                    owner_player_index = entry.owner.player_index_in_bounds(len(runtime_state.shots_hit))
                    if owner_player_index is not None and creature_lifecycle_is_alive(
                        creatures[int(hit_idx)].lifecycle_stage,
                    ):
                        shots_hit = runtime_state.shots_hit
                        shots_hit[owner_player_index] += 1

                if sfx_queue is not None:
                    sfx_queue.append(SfxId.EXPLOSION_MEDIUM)

                det_scale = 0.5
                damage_speed_mul = 0.0
                damage_base = 150.0
                burst_scale: float | None = None
                burst_min_detail = 0
                extra_decals = 0
                extra_radius = 0.0
                freeze_shard_target_pos = False
                match rule:
                    case RocketRule(
                        detonation_scale=detonation_scale,
                        damage_speed_mul=rule_damage_speed_mul,
                        damage_base=rule_damage_base,
                        burst_scale=rule_burst_scale,
                        burst_min_detail=rule_burst_min_detail,
                        extra_decals=rule_extra_decals,
                        extra_radius=rule_extra_radius,
                        freeze_shard_target_pos=rule_freeze_shard_target_pos,
                    ):
                        det_scale = float(detonation_scale)
                        damage_speed_mul = float(rule_damage_speed_mul)
                        damage_base = float(rule_damage_base)
                        burst_scale = None if rule_burst_scale is None else float(rule_burst_scale)
                        burst_min_detail = int(rule_burst_min_detail)
                        extra_decals = int(rule_extra_decals)
                        extra_radius = float(rule_extra_radius)
                        freeze_shard_target_pos = bool(rule_freeze_shard_target_pos)
                    case HomingRocketRule(
                        detonation_scale=detonation_scale,
                        damage_speed_mul=rule_damage_speed_mul,
                        damage_base=rule_damage_base,
                        extra_decals=rule_extra_decals,
                        extra_radius=rule_extra_radius,
                        freeze_shard_target_pos=rule_freeze_shard_target_pos,
                    ):
                        det_scale = float(detonation_scale)
                        damage_speed_mul = float(rule_damage_speed_mul)
                        damage_base = float(rule_damage_base)
                        extra_decals = int(rule_extra_decals)
                        extra_radius = float(rule_extra_radius)
                        freeze_shard_target_pos = bool(rule_freeze_shard_target_pos)
                    case RocketMinigunRule(
                        detonation_scale=detonation_scale,
                        damage_speed_mul=rule_damage_speed_mul,
                        damage_base=rule_damage_base,
                        extra_decals=rule_extra_decals,
                        extra_radius=rule_extra_radius,
                        freeze_shard_target_pos=rule_freeze_shard_target_pos,
                    ):
                        det_scale = float(detonation_scale)
                        damage_speed_mul = float(rule_damage_speed_mul)
                        damage_base = float(rule_damage_base)
                        extra_decals = int(rule_extra_decals)
                        extra_radius = float(rule_extra_radius)
                        freeze_shard_target_pos = bool(rule_freeze_shard_target_pos)

                if freeze_active:
                    if effects is not None:
                        for _ in range(4):
                            shard_angle = float(
                                rng.rand(caller=RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_PRE_HIT_FREEZE_SHARD_ANGLE)
                                % 612,
                            ) * 0.01
                            effects.spawn_freeze_shard(
                                pos=entry.pos,
                                angle=shard_angle,
                                rng=rng,
                                detail_preset=int(detail_preset),
                            )
                elif fx_queue is not None:
                    for dx_caller, dy_caller in _SECONDARY_PRE_HIT_DECAL_CALLERS:
                        offset = Vec2(
                            float(rng.rand(caller=dx_caller) % 20 - 10),
                            float(rng.rand(caller=dy_caller) % 20 - 10),
                        )
                        fx_queue.add_random(
                            pos=creatures[hit_idx].pos + offset,
                            rng=rng,
                        )

                if burst_scale is not None and effects is not None and int(detail_preset) > int(burst_min_detail):
                    effects.spawn_explosion_burst(
                        pos=entry.pos,
                        scale=float(burst_scale),
                        rng=rng,
                        detail_preset=int(detail_preset),
                    )

                # Native `projectile_update` applies hit visuals before
                # `creature_apply_damage` for secondary projectiles.
                damage = entry.speed * float(damage_speed_mul) + float(damage_base)
                _apply_secondary_damage(
                    hit_idx,
                    damage,
                    owner=entry.owner,
                    impulse=entry.vel / float(dt),
                )
                creature_spatial.sync_index(int(hit_idx))

                entry.type_id = SecondaryProjectileTypeId.DETONATION
                entry.vel = Vec2()
                entry.detonation_t = 0.0
                entry.detonation_scale = float(det_scale)
                entry.trail_timer = 0.0

                # Extra debris/scorch decals (or freeze shards) on detonation.
                if freeze_active:
                    if effects is not None:
                        shard_pos = entry.pos
                        freeze_angle_caller = RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_ROCKET_FREEZE_SHARD_ANGLE
                        if isinstance(rule, HomingRocketRule):
                            freeze_angle_caller = (
                                RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_SEEKER_ROCKET_FREEZE_SHARD_ANGLE
                            )
                        elif isinstance(rule, RocketMinigunRule):
                            freeze_angle_caller = (
                                RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_ROCKET_MINIGUN_FREEZE_SHARD_ANGLE
                            )
                        if freeze_shard_target_pos:
                            shard_pos = creatures[hit_idx].pos
                        for _ in range(8):
                            shard_angle = float(rng.rand(caller=freeze_angle_caller) % 612) * 0.01
                            effects.spawn_freeze_shard(
                                pos=shard_pos,
                                angle=shard_angle,
                                rng=rng,
                                detail_preset=int(detail_preset),
                            )
                else:
                    if fx_queue is not None and extra_decals > 0:
                        center = creatures[hit_idx].pos
                        angle_caller = RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_ROCKET_DECAL_ANGLE
                        radius_caller = RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_ROCKET_DECAL_RADIUS
                        if isinstance(rule, HomingRocketRule):
                            angle_caller = RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_SEEKER_ROCKET_DECAL_ANGLE
                            radius_caller = RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_SEEKER_ROCKET_DECAL_RADIUS
                        elif isinstance(rule, RocketMinigunRule):
                            angle_caller = RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_ROCKET_MINIGUN_DECAL_ANGLE
                            radius_caller = RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_ROCKET_MINIGUN_DECAL_RADIUS
                        for _ in range(int(extra_decals)):
                            angle = float(rng.rand(caller=angle_caller) % 628) * 0.01
                            if isinstance(rule, HomingRocketRule):
                                radius = float(rng.rand(caller=radius_caller) & 0x3F)
                            else:
                                radius = float(rng.rand(caller=radius_caller) % max(1, int(extra_radius)))
                            fx_queue.add_random(
                                pos=center + Vec2.from_angle(angle) * radius,
                                rng=rng,
                            )

                if sprite_effects is not None:
                    step = math.tau / 10.0
                    for idx in range(10):
                        mag = float(
                            rng.rand(caller=RngCallerStatic.SECONDARY_PROJECTILE_UPDATE_DETONATION_SPRITE_MAG) % 800,
                        ) * 0.1
                        ang = float(idx) * step
                        velocity = Vec2.from_angle(ang) * mag
                        sprite_effects.spawn(
                            pos=entry.pos,
                            vel=velocity,
                            scale=14.0,
                            color=RGBA(1.0, 1.0, 1.0, 0.37),
                        )

                continue

            if entry.speed < 0.0:
                entry.type_id = SecondaryProjectileTypeId.DETONATION
                entry.vel = Vec2()
                entry.detonation_t = 0.0
                entry.detonation_scale = 0.5
                entry.trail_timer = 0.0



from __future__ import annotations

from collections.abc import Sequence
from enum import IntEnum
from pathlib import Path
from typing import Any

import msgspec
from construct import Array, Byte, Bytes, Float32l, Int32sl, Struct

from crimson.aim_schemes import AimScheme
from crimson.game_modes import GameMode
from crimson.movement_controls import MovementControlType
from crimson.quests.level import QuestLevel

CRIMSON_CFG_NAME = "crimson.cfg"
CRIMSON_CFG_SIZE = 0x480
PLAYER_NAME_SIZE = 0x20
PLAYER_NAME_MAX_BYTES = PLAYER_NAME_SIZE - 1
SAVED_NAME_SLOT_COUNT = 8
SAVED_NAME_ENTRY_SIZE = 0x1B
SAVED_NAMES_BLOB_SIZE = SAVED_NAME_SLOT_COUNT * SAVED_NAME_ENTRY_SIZE
UNKNOWN_248_SIZE = 0x1F8
PLAYER_BIND_BLOCK_DWORDS = 0x10
PLAYER_BIND_BLOCK_SIZE = PLAYER_BIND_BLOCK_DWORDS * 4
EXT_DIRECTION_ARROW_FLAG_COUNT = 2
EXTENDED_RESERVED_GAP_SIZE = UNKNOWN_248_SIZE - 2 * PLAYER_BIND_BLOCK_SIZE - EXT_DIRECTION_ARROW_FLAG_COUNT
EXT_DIRECTION_ARROW_UNSET = 0
EXT_DIRECTION_ARROW_OFF = 1
EXT_DIRECTION_ARROW_ON = 2
KEYBIND_UNBOUND_CODE = 0x17E
RESERVED_KEYBIND_SLOT_COUNT = 2
PADDING_KEYBIND_SLOT_COUNT = 3
_DEFAULT_WIRE_RESERVED_KEYS = (KEYBIND_UNBOUND_CODE, KEYBIND_UNBOUND_CODE)
_DEFAULT_WIRE_PADDING = (KEYBIND_UNBOUND_CODE, KEYBIND_UNBOUND_CODE, KEYBIND_UNBOUND_CODE)

PLAYER_BIND_BLOCK_STRUCT = Struct(
    "move_forward" / Int32sl,
    "move_backward" / Int32sl,
    "turn_left" / Int32sl,
    "turn_right" / Int32sl,
    "fire" / Int32sl,
    "reserved_keys" / Array(RESERVED_KEYBIND_SLOT_COUNT, Int32sl),
    "aim_left" / Int32sl,
    "aim_right" / Int32sl,
    "axis_aim_y" / Int32sl,
    "axis_aim_x" / Int32sl,
    "axis_move_y" / Int32sl,
    "axis_move_x" / Int32sl,
    "padding" / Array(PADDING_KEYBIND_SLOT_COUNT, Int32sl),
)

CRIMSON_CFG_STRUCT = Struct(
    "sound_disable" / Byte,
    "music_disable" / Byte,
    "highscore_date_mode" / Byte,
    "highscore_duplicate_mode" / Byte,
    "direction_arrow_flags" / Array(2, Byte),
    "unknown_06" / Bytes(2),
    "unknown_08" / Int32sl,
    "unknown_0c" / Bytes(2),
    "fx_detail_0" / Byte,
    "unknown_0f" / Byte,
    "fx_detail_1" / Byte,
    "fx_detail_2" / Byte,
    "unknown_12" / Bytes(2),
    "player_count" / Int32sl,
    "game_mode" / Int32sl,
    "player_mode_flags" / Array(4, Int32sl),
    "player_mode_flags_reserved" / Bytes(0x18),
    "aim_schemes" / Array(4, Int32sl),
    "aim_schemes_reserved" / Bytes(0x18),
    "unknown_6c" / Int32sl,
    "texture_scale" / Float32l,
    "player_name_buf" / Bytes(12),
    "selected_saved_name_slot" / Int32sl,
    "saved_name_count" / Int32sl,
    "saved_name_order" / Array(SAVED_NAME_SLOT_COUNT, Int32sl),
    "saved_names" / Bytes(SAVED_NAMES_BLOB_SIZE),
    "player_name" / Bytes(PLAYER_NAME_SIZE),
    "player_name_len" / Int32sl,
    "unknown_1a4" / Int32sl,
    "unknown_1a8" / Int32sl,
    "unknown_1ac" / Int32sl,
    "aim_pov_right" / Int32sl,
    "aim_pov_left" / Int32sl,
    "screen_bpp" / Int32sl,
    "screen_width" / Int32sl,
    "screen_height" / Int32sl,
    "windowed_flag" / Byte,
    "unknown_1c5" / Bytes(3),
    "keybinds_p1_p2" / Array(2, PLAYER_BIND_BLOCK_STRUCT),
    # The original wire format leaves this 0x1F8-byte gap uninterpreted.
    # The Python port uses the front of it for P3/P4 control bindings and
    # their direction-arrow flags, while preserving the remaining bytes.
    "extended_keybinds_p3_p4" / Array(2, PLAYER_BIND_BLOCK_STRUCT),
    "extended_direction_arrow_flags" / Array(EXT_DIRECTION_ARROW_FLAG_COUNT, Byte),
    "extended_reserved_gap" / Bytes(EXTENDED_RESERVED_GAP_SIZE),
    "unknown_440" / Int32sl,
    "unknown_444" / Int32sl,
    "hardcore_flag" / Byte,
    "ui_info_texts" / Byte,
    "unknown_44a" / Bytes(2),
    "perk_prompt_counter" / Int32sl,
    "unknown_450" / Int32sl,
    "unknown_454" / Bytes(0x0C),
    "sound_freq_adjustment_enabled" / Byte,
    "unknown_461" / Bytes(3),
    "sfx_volume" / Float32l,
    "music_volume" / Float32l,
    "violence_disabled" / Byte,
    "score_load_gate" / Byte,
    "safe_mode_backend_enabled" / Byte,
    "unknown_46f" / Byte,
    "detail_preset" / Int32sl,
    "mouse_sensitivity" / Float32l,
    "keybind_pick_perk" / Int32sl,
    "keybind_reload" / Int32sl,
)

_DEFAULT_PROFILE_NAME = "10tons"
_DEFAULT_SAVED_NAMES: tuple[str, str, str, str, str, str, str, str] = (
    "default",
    "default",
    "default",
    "default",
    "default",
    "default",
    "default",
    "default",
)


class HighScoreDateMode(IntEnum):
    ALL_TIME = 0
    MONTH = 1
    WEEK = 2
    DAY = 3


class CrimsonDisplayConfig(msgspec.Struct):
    width: int
    height: int
    windowed: bool
    bpp: int
    texture_scale: float
    mouse_sensitivity: float
    detail_preset: int
    fx_detail: tuple[bool, bool, bool]
    violence_disabled: int

    def fx_detail_enabled(self, level: int, default: bool = False) -> bool:
        _ = default
        return bool(self.fx_detail[int(level)])

    def set_fx_detail(self, level: int, enabled: bool) -> None:
        idx = int(level)
        values = list(self.fx_detail)
        values[idx] = bool(enabled)
        self.fx_detail = (bool(values[0]), bool(values[1]), bool(values[2]))


class CrimsonAudioConfig(msgspec.Struct):
    sound_disabled: bool
    music_disabled: bool
    sfx_volume: float
    music_volume: float


class CrimsonGameplayConfig(msgspec.Struct):
    mode: GameMode
    player_count: int
    hardcore: bool
    quest_level: QuestLevel | None
    show_info_texts: bool


class CrimsonProfileConfig(msgspec.Struct):
    player_name: str
    player_name_input_len: int
    saved_name_count: int
    selected_saved_name_slot: int
    saved_names: tuple[str, str, str, str, str, str, str, str]
    show_internet_scores: bool
    score_date_mode: HighScoreDateMode

    def set_player_name_input(self, name: str) -> None:
        encoded = str(name).encode("latin-1", errors="ignore")[:PLAYER_NAME_MAX_BYTES]
        buf = bytearray(PLAYER_NAME_SIZE)
        buf[: len(encoded)] = encoded
        buf[min(len(encoded), PLAYER_NAME_MAX_BYTES)] = 0

        end = buf.index(0)
        i = end - 1
        while i > 0 and buf[i] == 0x20:
            buf[i] = 0
            i -= 1

        self.player_name = bytes(buf).split(b"\x00", 1)[0].decode("latin-1", errors="ignore")
        self.player_name_input_len = len(encoded)

    def saved_name_labels(self) -> tuple[str, ...]:
        count = int(self.saved_name_count)
        if count < 1 or count > SAVED_NAME_SLOT_COUNT:
            raise ValueError(f"saved_name_count must be in 1..{SAVED_NAME_SLOT_COUNT}, got {count}")
        labels: list[str] = []
        for idx in range(count):
            label = str(self.saved_names[idx]).strip()
            if not label:
                label = "default" if idx == 0 else f"slot_{idx}"
            labels.append(label)
        return tuple(labels)


class CrimsonPlayerControls(msgspec.Struct):
    movement: MovementControlType
    aim_scheme: AimScheme
    show_direction_arrow: bool
    move_codes: tuple[int, int, int, int]
    fire_code: int
    keyboard_aim_codes: tuple[int, int]
    aim_axis_codes: tuple[int, int]
    move_axis_codes: tuple[int, int]


_DEFAULT_PLAYER_CONTROL_TEMPLATES: tuple[CrimsonPlayerControls, ...] = (
    CrimsonPlayerControls(
        movement=MovementControlType.STATIC,
        aim_scheme=AimScheme.MOUSE,
        show_direction_arrow=True,
        move_codes=(0x11, 0x1F, 0x1E, 0x20),
        fire_code=0x100,
        keyboard_aim_codes=(0x10, 0x12),
        aim_axis_codes=(0x13F, 0x140),
        move_axis_codes=(0x141, 0x153),
    ),
    CrimsonPlayerControls(
        movement=MovementControlType.STATIC,
        aim_scheme=AimScheme.MOUSE,
        show_direction_arrow=True,
        move_codes=(0xC8, 0xD0, 0xCB, 0xCD),
        fire_code=0x9D,
        keyboard_aim_codes=(0xD3, 0xD1),
        aim_axis_codes=(0x13F, 0x140),
        move_axis_codes=(0x141, 0x153),
    ),
    CrimsonPlayerControls(
        movement=MovementControlType.STATIC,
        aim_scheme=AimScheme.MOUSE,
        show_direction_arrow=True,
        move_codes=(0x17, 0x25, 0x24, 0x26),
        fire_code=0x36,
        keyboard_aim_codes=(0x16, 0x18),
        aim_axis_codes=(0x17E, 0x17E),
        move_axis_codes=(0x17E, 0x17E),
    ),
    CrimsonPlayerControls(
        movement=MovementControlType.STATIC,
        aim_scheme=AimScheme.MOUSE,
        show_direction_arrow=True,
        move_codes=(0x131, 0x132, 0x133, 0x134),
        fire_code=0x11F,
        keyboard_aim_codes=(0x17E, 0x17E),
        aim_axis_codes=(0x140, 0x13F),
        move_axis_codes=(0x153, 0x154),
    ),
)


class CrimsonControlsConfig(msgspec.Struct):
    players: tuple[CrimsonPlayerControls, CrimsonPlayerControls, CrimsonPlayerControls, CrimsonPlayerControls]
    pick_perk_code: int
    reload_code: int

    def player(self, player_index: int) -> CrimsonPlayerControls:
        return self.players[_player_index(player_index)]


class CrimsonConfig(msgspec.Struct):
    path: Path
    display: CrimsonDisplayConfig
    audio: CrimsonAudioConfig
    gameplay: CrimsonGameplayConfig
    profile: CrimsonProfileConfig
    controls: CrimsonControlsConfig

    def save(self) -> None:
        self.path.write_bytes(encode_crimson_cfg(self))


def _player_index(player_index: int) -> int:
    idx = int(player_index)
    if idx < 0 or idx >= 4:
        raise IndexError(f"player index must be in 0..3, got {idx}")
    return idx


def _require_range(value: int, *, minimum: int, maximum: int, field: str) -> int:
    if value < minimum or value > maximum:
        raise ValueError(f"{field} must be in {minimum}..{maximum}, got {value}")
    return value


def _parsed_player_bind_block(raw: dict[str, Any], *, player_index: int) -> dict[str, Any]:
    idx = _player_index(player_index)
    if idx < 2:
        return raw["keybinds_p1_p2"][idx]
    return raw["extended_keybinds_p3_p4"][idx - 2]


def _parsed_player_bind_block_is_uninitialized(raw_block: dict[str, Any]) -> bool:
    return not any(
        (
            raw_block["move_forward"],
            raw_block["move_backward"],
            raw_block["turn_left"],
            raw_block["turn_right"],
            raw_block["fire"],
            *raw_block["reserved_keys"],
            raw_block["aim_left"],
            raw_block["aim_right"],
            raw_block["axis_aim_y"],
            raw_block["axis_aim_x"],
            raw_block["axis_move_y"],
            raw_block["axis_move_x"],
        ),
    )


def _player_controls_from_parsed_bind_block(
    raw_block: dict[str, Any],
    *,
    player_index: int,
    movement: MovementControlType,
    aim_scheme: AimScheme,
    show_direction_arrow: bool,
) -> CrimsonPlayerControls:
    if _parsed_player_bind_block_is_uninitialized(raw_block):
        defaults = _default_player_controls(player_index)
        return CrimsonPlayerControls(
            movement=movement,
            aim_scheme=aim_scheme,
            show_direction_arrow=show_direction_arrow,
            move_codes=defaults.move_codes,
            fire_code=defaults.fire_code,
            keyboard_aim_codes=defaults.keyboard_aim_codes,
            aim_axis_codes=defaults.aim_axis_codes,
            move_axis_codes=defaults.move_axis_codes,
        )
    return CrimsonPlayerControls(
        movement=movement,
        aim_scheme=aim_scheme,
        show_direction_arrow=show_direction_arrow,
        move_codes=(
            raw_block["move_forward"],
            raw_block["move_backward"],
            raw_block["turn_left"],
            raw_block["turn_right"],
        ),
        fire_code=raw_block["fire"],
        keyboard_aim_codes=(raw_block["aim_left"], raw_block["aim_right"]),
        aim_axis_codes=(raw_block["axis_aim_y"], raw_block["axis_aim_x"]),
        move_axis_codes=(raw_block["axis_move_y"], raw_block["axis_move_x"]),
    )


def _encode_player_bind_block(player: CrimsonPlayerControls) -> dict[str, object]:
    return {
        "move_forward": player.move_codes[0],
        "move_backward": player.move_codes[1],
        "turn_left": player.move_codes[2],
        "turn_right": player.move_codes[3],
        "fire": player.fire_code,
        "reserved_keys": [int(_DEFAULT_WIRE_RESERVED_KEYS[0]), int(_DEFAULT_WIRE_RESERVED_KEYS[1])],
        "aim_left": player.keyboard_aim_codes[0],
        "aim_right": player.keyboard_aim_codes[1],
        "axis_aim_y": player.aim_axis_codes[0],
        "axis_aim_x": player.aim_axis_codes[1],
        "axis_move_y": player.move_axis_codes[0],
        "axis_move_x": player.move_axis_codes[1],
        "padding": [int(_DEFAULT_WIRE_PADDING[0]), int(_DEFAULT_WIRE_PADDING[1]), int(_DEFAULT_WIRE_PADDING[2])],
    }


def _encode_primary_keybinds(players: Sequence[CrimsonPlayerControls]) -> list[dict[str, object]]:
    return [_encode_player_bind_block(players[idx]) for idx in range(2)]


def _encode_extended_keybinds(players: Sequence[CrimsonPlayerControls]) -> list[dict[str, object]]:
    return [_encode_player_bind_block(players[idx]) for idx in range(2, 4)]


def _decode_direction_arrow(raw: dict, *, player_index: int) -> bool:
    idx = _player_index(player_index)
    if idx < 2:
        return bool(int(raw["direction_arrow_flags"][idx]))

    value = int(raw["extended_direction_arrow_flags"][idx - 2])
    if value == EXT_DIRECTION_ARROW_OFF:
        return False
    if value in (EXT_DIRECTION_ARROW_UNSET, EXT_DIRECTION_ARROW_ON):
        return True
    raise ValueError(f"unsupported extended direction arrow flag value: {value}")


def _encode_direction_arrow_flags(players: Sequence[CrimsonPlayerControls]) -> tuple[list[int], tuple[int, int]]:
    primary = [1, 1]
    for idx in range(2):
        primary[idx] = 1 if bool(players[idx].show_direction_arrow) else 0
    extended: list[int] = []
    for idx in range(2, 4):
        extended.append(EXT_DIRECTION_ARROW_ON if bool(players[idx].show_direction_arrow) else EXT_DIRECTION_ARROW_OFF)
    return primary, (extended[0], extended[1])


def _decode_movement(value: int) -> MovementControlType:
    raw = int(value)
    if raw == 0:
        return MovementControlType.STATIC
    return MovementControlType(raw)


def _decode_aim_scheme(value: int) -> AimScheme:
    return AimScheme(int(value))


def _decode_game_mode(value: int) -> GameMode:
    return GameMode(int(value))


def _decode_high_score_date_mode(value: int) -> HighScoreDateMode:
    return HighScoreDateMode(int(value))


def _decode_player_name(raw: bytes) -> str:
    return bytes(raw).split(b"\x00", 1)[0].decode("latin-1", errors="ignore")


def _encode_player_name_buffer(name: str) -> bytes:
    encoded = str(name).encode("latin-1", errors="ignore")[:PLAYER_NAME_MAX_BYTES]
    buf = bytearray(PLAYER_NAME_SIZE)
    buf[: len(encoded)] = encoded
    buf[min(len(encoded), PLAYER_NAME_MAX_BYTES)] = 0
    return bytes(buf)


def _decode_saved_names(raw: bytes) -> tuple[str, str, str, str, str, str, str, str]:
    blob = bytes(raw)
    names: list[str] = []
    for idx in range(SAVED_NAME_SLOT_COUNT):
        entry = blob[idx * SAVED_NAME_ENTRY_SIZE : (idx + 1) * SAVED_NAME_ENTRY_SIZE]
        names.append(entry.split(b"\x00", 1)[0].decode("latin-1", errors="ignore"))
    return tuple(names)  # type: ignore[return-value]


def _encode_saved_names_blob(names: Sequence[str]) -> bytes:
    out = bytearray(SAVED_NAMES_BLOB_SIZE)
    for idx in range(SAVED_NAME_SLOT_COUNT):
        name = str(names[idx]) if idx < len(names) else ""
        encoded = name.encode("latin-1", errors="ignore")[: SAVED_NAME_ENTRY_SIZE - 1]
        start = idx * SAVED_NAME_ENTRY_SIZE
        out[start : start + SAVED_NAME_ENTRY_SIZE] = b"\x00" * SAVED_NAME_ENTRY_SIZE
        out[start : start + len(encoded)] = encoded
        out[start + min(len(encoded), SAVED_NAME_ENTRY_SIZE - 1)] = 0
    return bytes(out)


def _saved_name_order_values() -> tuple[int, ...]:
    return tuple(range(SAVED_NAME_SLOT_COUNT))


def _default_player_controls(player_index: int) -> CrimsonPlayerControls:
    defaults = _DEFAULT_PLAYER_CONTROL_TEMPLATES[_player_index(player_index)]
    return CrimsonPlayerControls(
        movement=defaults.movement,
        aim_scheme=defaults.aim_scheme,
        show_direction_arrow=defaults.show_direction_arrow,
        move_codes=defaults.move_codes,
        fire_code=defaults.fire_code,
        keyboard_aim_codes=defaults.keyboard_aim_codes,
        aim_axis_codes=defaults.aim_axis_codes,
        move_axis_codes=defaults.move_axis_codes,
    )


def default_crimson_cfg(path: Path = Path("<memory>")) -> CrimsonConfig:
    profile = CrimsonProfileConfig(
        player_name="",
        player_name_input_len=0,
        saved_name_count=1,
        selected_saved_name_slot=0,
        saved_names=_DEFAULT_SAVED_NAMES,
        show_internet_scores=False,
        score_date_mode=HighScoreDateMode.ALL_TIME,
    )
    profile.set_player_name_input(_DEFAULT_PROFILE_NAME)
    profile.player_name_input_len = 0
    return CrimsonConfig(
        path=path,
        display=CrimsonDisplayConfig(
            width=1024,
            height=768,
            windowed=True,
            bpp=32,
            texture_scale=1.0,
            mouse_sensitivity=0.5,
            detail_preset=5,
            fx_detail=(True, True, True),
            violence_disabled=0,
        ),
        audio=CrimsonAudioConfig(
            sound_disabled=False,
            music_disabled=False,
            sfx_volume=1.0,
            music_volume=1.0,
        ),
        gameplay=CrimsonGameplayConfig(
            mode=GameMode.SURVIVAL,
            player_count=1,
            hardcore=False,
            quest_level=None,
            show_info_texts=True,
        ),
        profile=profile,
        controls=CrimsonControlsConfig(
            players=(
                _default_player_controls(0),
                _default_player_controls(1),
                _default_player_controls(2),
                _default_player_controls(3),
            ),
            pick_perk_code=0x101,
            reload_code=0x102,
        ),
    )


def decode_crimson_cfg(path: Path, blob: bytes) -> CrimsonConfig:
    if len(blob) != CRIMSON_CFG_SIZE:
        raise ValueError(f"{path} has unexpected size {len(blob)} (expected {CRIMSON_CFG_SIZE})")
    raw = CRIMSON_CFG_STRUCT.parse(blob)

    fx_detail = (
        bool(raw["fx_detail_0"]),
        bool(raw["fx_detail_1"]),
        bool(raw["fx_detail_2"]),
    )
    detail_preset = int(raw["detail_preset"])
    if detail_preset == 0 and not any(fx_detail):
        detail_preset = 5
        fx_detail = (True, True, True)
    else:
        detail_preset = _require_range(detail_preset, minimum=1, maximum=5, field="detail_preset")

    players = tuple(
        _player_controls_from_parsed_bind_block(
            _parsed_player_bind_block(raw, player_index=idx),
            player_index=idx,
            movement=_decode_movement(raw["player_mode_flags"][idx]),
            aim_scheme=_decode_aim_scheme(raw["aim_schemes"][idx]),
            show_direction_arrow=_decode_direction_arrow(raw, player_index=idx),
        )
        for idx in range(4)
    )

    return CrimsonConfig(
        path=path,
        display=CrimsonDisplayConfig(
            width=int(raw["screen_width"]),
            height=int(raw["screen_height"]),
            windowed=bool(raw["windowed_flag"]),
            bpp=int(raw["screen_bpp"]),
            texture_scale=float(raw["texture_scale"]),
            mouse_sensitivity=float(raw["mouse_sensitivity"]),
            detail_preset=detail_preset,
            fx_detail=fx_detail,
            violence_disabled=int(raw["violence_disabled"]),
        ),
        audio=CrimsonAudioConfig(
            sound_disabled=bool(raw["sound_disable"]),
            music_disabled=bool(raw["music_disable"]),
            sfx_volume=float(raw["sfx_volume"]),
            music_volume=float(raw["music_volume"]),
        ),
        gameplay=CrimsonGameplayConfig(
            mode=_decode_game_mode(raw["game_mode"]),
            player_count=_require_range(int(raw["player_count"]), minimum=1, maximum=4, field="player_count"),
            hardcore=bool(raw["hardcore_flag"]),
            quest_level=None,
            show_info_texts=bool(raw["ui_info_texts"]),
        ),
        profile=CrimsonProfileConfig(
            player_name=_decode_player_name(raw["player_name"]),
            player_name_input_len=_require_range(
                int(raw["player_name_len"]),
                minimum=0,
                maximum=PLAYER_NAME_MAX_BYTES,
                field="player_name_len",
            ),
            saved_name_count=_require_range(
                int(raw["saved_name_count"]),
                minimum=1,
                maximum=SAVED_NAME_SLOT_COUNT,
                field="saved_name_count",
            ),
            selected_saved_name_slot=_require_range(
                int(raw["selected_saved_name_slot"]),
                minimum=0,
                maximum=SAVED_NAME_SLOT_COUNT - 1,
                field="selected_saved_name_slot",
            ),
            saved_names=_decode_saved_names(raw["saved_names"]),
            show_internet_scores=bool(raw["score_load_gate"]),
            score_date_mode=_decode_high_score_date_mode(raw["highscore_date_mode"]),
        ),
        controls=CrimsonControlsConfig(
            players=players,  # type: ignore[arg-type]
            pick_perk_code=int(raw["keybind_pick_perk"]),
            reload_code=int(raw["keybind_reload"]),
        ),
    )


def _canonical_wire_data() -> dict:
    return dict(CRIMSON_CFG_STRUCT.parse(bytes(CRIMSON_CFG_SIZE)))


def encode_crimson_cfg(config: CrimsonConfig) -> bytes:
    data = _canonical_wire_data()
    primary_direction_arrows, extended_direction_arrows = _encode_direction_arrow_flags(config.controls.players)

    data["sound_disable"] = 1 if config.audio.sound_disabled else 0
    data["music_disable"] = 1 if config.audio.music_disabled else 0
    data["highscore_date_mode"] = int(config.profile.score_date_mode)
    data["direction_arrow_flags"] = primary_direction_arrows
    data["unknown_08"] = 0
    data["fx_detail_0"] = 1 if config.display.fx_detail_enabled(0) else 0
    data["fx_detail_1"] = 1 if config.display.fx_detail_enabled(1) else 0
    data["fx_detail_2"] = 1 if config.display.fx_detail_enabled(2) else 0
    data["player_count"] = _require_range(
        int(config.gameplay.player_count),
        minimum=1,
        maximum=4,
        field="player_count",
    )
    data["game_mode"] = int(config.gameplay.mode)
    data["player_mode_flags"] = [int(player.movement) for player in config.controls.players]
    data["aim_schemes"] = [int(player.aim_scheme) for player in config.controls.players]
    data["texture_scale"] = float(config.display.texture_scale)
    data["selected_saved_name_slot"] = _require_range(
        int(config.profile.selected_saved_name_slot),
        minimum=0,
        maximum=SAVED_NAME_SLOT_COUNT - 1,
        field="selected_saved_name_slot",
    )
    data["saved_name_count"] = _require_range(
        int(config.profile.saved_name_count),
        minimum=1,
        maximum=SAVED_NAME_SLOT_COUNT,
        field="saved_name_count",
    )
    data["saved_name_order"] = list(_saved_name_order_values())
    data["saved_names"] = _encode_saved_names_blob(config.profile.saved_names)
    data["player_name"] = _encode_player_name_buffer(config.profile.player_name)
    data["player_name_len"] = _require_range(
        int(config.profile.player_name_input_len),
        minimum=0,
        maximum=PLAYER_NAME_MAX_BYTES,
        field="player_name_len",
    )
    data["unknown_1a4"] = 100
    data["aim_pov_right"] = 9000
    data["aim_pov_left"] = 27000
    data["screen_bpp"] = int(config.display.bpp)
    data["screen_width"] = int(config.display.width)
    data["screen_height"] = int(config.display.height)
    data["windowed_flag"] = 1 if config.display.windowed else 0
    data["keybinds_p1_p2"] = _encode_primary_keybinds(config.controls.players)
    data["extended_keybinds_p3_p4"] = _encode_extended_keybinds(config.controls.players)
    data["extended_direction_arrow_flags"] = [int(extended_direction_arrows[0]), int(extended_direction_arrows[1])]
    data["hardcore_flag"] = 1 if config.gameplay.hardcore else 0
    data["ui_info_texts"] = 1 if config.gameplay.show_info_texts else 0
    data["unknown_450"] = 1
    data["sound_freq_adjustment_enabled"] = 1
    data["sfx_volume"] = float(config.audio.sfx_volume)
    data["music_volume"] = float(config.audio.music_volume)
    data["violence_disabled"] = int(config.display.violence_disabled)
    data["score_load_gate"] = 1 if config.profile.show_internet_scores else 0
    data["detail_preset"] = _require_range(
        int(config.display.detail_preset),
        minimum=1,
        maximum=5,
        field="detail_preset",
    )
    data["mouse_sensitivity"] = float(config.display.mouse_sensitivity)
    data["keybind_pick_perk"] = config.controls.pick_perk_code
    data["keybind_reload"] = config.controls.reload_code
    return CRIMSON_CFG_STRUCT.build(data)


def load_crimson_cfg(path: Path) -> CrimsonConfig:
    return decode_crimson_cfg(path, path.read_bytes())


def ensure_crimson_cfg(base_dir: Path) -> CrimsonConfig:
    path = base_dir / CRIMSON_CFG_NAME
    if not path.exists():
        config = default_crimson_cfg(path)
        config.save()
        return config
    return load_crimson_cfg(path)


def apply_detail_preset(config: CrimsonConfig, preset: int | None = None) -> int:
    selected = config.display.detail_preset if preset is None else int(preset)
    selected = _require_range(selected, minimum=1, maximum=5, field="detail_preset")
    config.display.detail_preset = selected
    if selected <= 1:
        config.display.fx_detail = (False, False, False)
    elif selected == 2:
        config.display.fx_detail = (False, False, True)
    else:
        config.display.fx_detail = (True, True, True)
    return selected



from __future__ import annotations

import io
from enum import Enum, auto
from pathlib import Path
from typing import TYPE_CHECKING, Final, cast

import msgspec
from PIL import Image

from grim.raylib_api import rl

from . import jaz, paq

if TYPE_CHECKING:
    from grim.fonts.small import SmallFontData

PAQ_NAME = "crimson.paq"


class TextureId(Enum):
    BACKPLASMA = auto()
    MOCKUP = auto()
    LOGO_ESRB = auto()
    LOADING = auto()
    CL_LOGO = auto()
    SPLASH_10TONS = auto()
    SPLASH_REFLEXIVE = auto()
    DEFAULT_FONT_COURIER = auto()
    SMALL_WHITE = auto()
    TROOPER = auto()
    ZOMBIE = auto()
    SPIDER_SP1 = auto()
    SPIDER_SP2 = auto()
    ALIEN = auto()
    LIZARD = auto()
    ARROW = auto()
    BULLET_I = auto()
    BULLET_TRAIL = auto()
    BODYSET = auto()
    PROJS = auto()
    UI_ICON_AIM = auto()
    UI_BUTTON_SM = auto()
    UI_BUTTON_MD = auto()
    UI_CHECK_ON = auto()
    UI_CHECK_OFF = auto()
    UI_RECT_OFF = auto()
    UI_RECT_ON = auto()
    BONUSES = auto()
    UI_IND_BULLET = auto()
    UI_IND_ROCKET = auto()
    UI_IND_ELECTRIC = auto()
    UI_IND_FIRE = auto()
    PARTICLES = auto()
    UI_IND_LIFE = auto()
    UI_IND_PANEL = auto()
    UI_ARROW = auto()
    UI_CURSOR = auto()
    UI_AIM = auto()
    TER_Q1_BASE = auto()
    TER_Q1_OVERLAY = auto()
    TER_Q2_BASE = auto()
    TER_Q2_OVERLAY = auto()
    TER_Q3_BASE = auto()
    TER_Q3_OVERLAY = auto()
    TER_Q4_BASE = auto()
    TER_Q4_OVERLAY = auto()
    UI_TEXT_LEVEL_COMPLETE = auto()
    UI_TEXT_QUEST = auto()
    UI_NUM1 = auto()
    UI_NUM2 = auto()
    UI_NUM3 = auto()
    UI_NUM4 = auto()
    UI_NUM5 = auto()
    UI_WICONS = auto()
    UI_GAME_TOP = auto()
    UI_LIFE_HEART = auto()
    UI_CLOCK_TABLE = auto()
    UI_CLOCK_POINTER = auto()
    MUZZLE_FLASH = auto()
    UI_DROP_ON = auto()
    UI_DROP_OFF = auto()
    UI_SIGN_CRIMSON = auto()
    UI_MENU_ITEM = auto()
    UI_MENU_PANEL = auto()
    UI_ITEM_TEXTS = auto()
    UI_TEXT_REAPER = auto()
    UI_TEXT_WELL_DONE = auto()
    UI_TEXT_CONTROLS = auto()
    UI_TEXT_PICK_A_PERK = auto()
    UI_TEXT_LEVEL_UP = auto()


class TextureSpec(msgspec.Struct, frozen=True):
    rel_path: str
    clamp: bool = False
    point_filter: bool = False


TEXTURE_SPECS: Final[dict[TextureId, TextureSpec]] = {
    TextureId.BACKPLASMA: TextureSpec("load/backplasma.jaz"),
    TextureId.MOCKUP: TextureSpec("load/mockup.jaz"),
    TextureId.LOGO_ESRB: TextureSpec("load/esrb_mature.jaz"),
    TextureId.LOADING: TextureSpec("load/loading.jaz"),
    TextureId.CL_LOGO: TextureSpec("load/logo_crimsonland.tga"),
    TextureId.SPLASH_10TONS: TextureSpec("load/splash10tons.jaz"),
    TextureId.SPLASH_REFLEXIVE: TextureSpec("load/splashReflexive.jpg"),
    TextureId.DEFAULT_FONT_COURIER: TextureSpec("load/default_font_courier.tga"),
    TextureId.SMALL_WHITE: TextureSpec("load/smallWhite.tga", point_filter=True),
    TextureId.TROOPER: TextureSpec("game/trooper.jaz"),
    TextureId.ZOMBIE: TextureSpec("game/zombie.jaz"),
    TextureId.SPIDER_SP1: TextureSpec("game/spider_sp1.jaz"),
    TextureId.SPIDER_SP2: TextureSpec("game/spider_sp2.jaz"),
    TextureId.ALIEN: TextureSpec("game/alien.jaz"),
    TextureId.LIZARD: TextureSpec("game/lizard.jaz"),
    TextureId.ARROW: TextureSpec("load/arrow.tga"),
    TextureId.BULLET_I: TextureSpec("load/bullet16.tga"),
    TextureId.BULLET_TRAIL: TextureSpec("load/bulletTrail.tga"),
    TextureId.BODYSET: TextureSpec("game/bodyset.jaz"),
    TextureId.PROJS: TextureSpec("game/projs.jaz"),
    TextureId.UI_ICON_AIM: TextureSpec("ui/ui_iconAim.jaz", clamp=True),
    TextureId.UI_BUTTON_SM: TextureSpec("ui/ui_button_64x32.jaz", clamp=True),
    TextureId.UI_BUTTON_MD: TextureSpec("ui/ui_button_128x32.jaz", clamp=True),
    TextureId.UI_CHECK_ON: TextureSpec("ui/ui_checkOn.jaz", clamp=True),
    TextureId.UI_CHECK_OFF: TextureSpec("ui/ui_checkOff.jaz", clamp=True),
    TextureId.UI_RECT_OFF: TextureSpec("ui/ui_rectOff.jaz", clamp=True),
    TextureId.UI_RECT_ON: TextureSpec("ui/ui_rectOn.jaz", clamp=True),
    TextureId.BONUSES: TextureSpec("game/bonuses.jaz"),
    TextureId.UI_IND_BULLET: TextureSpec("ui/ui_indBullet.jaz", clamp=True),
    TextureId.UI_IND_ROCKET: TextureSpec("ui/ui_indRocket.jaz", clamp=True),
    TextureId.UI_IND_ELECTRIC: TextureSpec("ui/ui_indElectric.jaz", clamp=True),
    TextureId.UI_IND_FIRE: TextureSpec("ui/ui_indFire.jaz", clamp=True),
    TextureId.PARTICLES: TextureSpec("game/particles.jaz"),
    TextureId.UI_IND_LIFE: TextureSpec("ui/ui_indLife.jaz", clamp=True),
    TextureId.UI_IND_PANEL: TextureSpec("ui/ui_indPanel.jaz", clamp=True),
    TextureId.UI_ARROW: TextureSpec("ui/ui_arrow.jaz", clamp=True),
    TextureId.UI_CURSOR: TextureSpec("ui/ui_cursor.jaz", clamp=True),
    TextureId.UI_AIM: TextureSpec("ui/ui_aim.jaz", clamp=True),
    TextureId.TER_Q1_BASE: TextureSpec("ter/ter_q1_base.jaz"),
    TextureId.TER_Q1_OVERLAY: TextureSpec("ter/ter_q1_tex1.jaz"),
    TextureId.TER_Q2_BASE: TextureSpec("ter/ter_q2_base.jaz"),
    TextureId.TER_Q2_OVERLAY: TextureSpec("ter/ter_q2_tex1.jaz"),
    TextureId.TER_Q3_BASE: TextureSpec("ter/ter_q3_base.jaz"),
    TextureId.TER_Q3_OVERLAY: TextureSpec("ter/ter_q3_tex1.jaz"),
    TextureId.TER_Q4_BASE: TextureSpec("ter/ter_q4_base.jaz"),
    TextureId.TER_Q4_OVERLAY: TextureSpec("ter/ter_q4_tex1.jaz"),
    TextureId.UI_TEXT_LEVEL_COMPLETE: TextureSpec("ui/ui_textLevComp.jaz", clamp=True),
    TextureId.UI_TEXT_QUEST: TextureSpec("ui/ui_textQuest.jaz", clamp=True),
    TextureId.UI_NUM1: TextureSpec("ui/ui_num1.jaz", clamp=True),
    TextureId.UI_NUM2: TextureSpec("ui/ui_num2.jaz", clamp=True),
    TextureId.UI_NUM3: TextureSpec("ui/ui_num3.jaz", clamp=True),
    TextureId.UI_NUM4: TextureSpec("ui/ui_num4.jaz", clamp=True),
    TextureId.UI_NUM5: TextureSpec("ui/ui_num5.jaz", clamp=True),
    TextureId.UI_WICONS: TextureSpec("ui/ui_wicons.jaz", clamp=True),
    TextureId.UI_GAME_TOP: TextureSpec("ui/ui_gameTop.jaz", clamp=True),
    TextureId.UI_LIFE_HEART: TextureSpec("ui/ui_lifeHeart.jaz", clamp=True),
    TextureId.UI_CLOCK_TABLE: TextureSpec("ui/ui_clockTable.jaz", clamp=True),
    TextureId.UI_CLOCK_POINTER: TextureSpec("ui/ui_clockPointer.jaz", clamp=True),
    TextureId.MUZZLE_FLASH: TextureSpec("game/muzzleFlash.jaz"),
    TextureId.UI_DROP_ON: TextureSpec("ui/ui_dropDownOn.jaz", clamp=True),
    TextureId.UI_DROP_OFF: TextureSpec("ui/ui_dropDownOff.jaz", clamp=True),
    TextureId.UI_SIGN_CRIMSON: TextureSpec("ui/ui_signCrimson.jaz", clamp=True),
    TextureId.UI_MENU_ITEM: TextureSpec("ui/ui_menuItem.jaz", clamp=True),
    TextureId.UI_MENU_PANEL: TextureSpec("ui/ui_menuPanel.jaz", clamp=True),
    TextureId.UI_ITEM_TEXTS: TextureSpec("ui/ui_itemTexts.jaz", clamp=True),
    TextureId.UI_TEXT_REAPER: TextureSpec("ui/ui_textReaper.jaz", clamp=True),
    TextureId.UI_TEXT_WELL_DONE: TextureSpec("ui/ui_textWellDone.jaz", clamp=True),
    TextureId.UI_TEXT_CONTROLS: TextureSpec("ui/ui_textControls.jaz", clamp=True),
    TextureId.UI_TEXT_PICK_A_PERK: TextureSpec("ui/ui_textPickAPerk.jaz", clamp=True),
    TextureId.UI_TEXT_LEVEL_UP: TextureSpec("ui/ui_textLevelUp.jaz", clamp=True),
}

class RuntimeResources(msgspec.Struct):
    assets_dir: Path
    textures: dict[TextureId, rl.Texture]
    small_font: SmallFontData

    def texture(self, texture_id: TextureId) -> rl.Texture:
        texture = self.textures.get(texture_id)
        if texture is None:
            rel_path = TEXTURE_SPECS[texture_id].rel_path
            raise RuntimeError(f"runtime texture is not available: {rel_path}")
        return texture

    def unload(self) -> None:
        seen: set[int] = set()
        for texture in self.textures.values():
            texture_id = int(texture.id)
            if texture_id <= 0 or texture_id in seen:
                continue
            rl.unload_texture(texture)
            seen.add(texture_id)
        self.textures.clear()


_REGISTERED_RESOURCES: dict[Path, RuntimeResources] = {}


def _normalize_assets_dir(assets_dir: Path) -> Path:
    return Path(assets_dir).resolve()


def register_runtime_resources(resources: RuntimeResources) -> None:
    _REGISTERED_RESOURCES[_normalize_assets_dir(resources.assets_dir)] = resources


def unregister_runtime_resources(assets_dir: Path) -> None:
    _REGISTERED_RESOURCES.pop(_normalize_assets_dir(assets_dir), None)


def runtime_resources_for(assets_dir: Path) -> RuntimeResources:
    normalized = _normalize_assets_dir(assets_dir)
    resources = _REGISTERED_RESOURCES.get(normalized)
    if resources is None:
        raise RuntimeError(f"runtime resources not loaded for {assets_dir}")
    return resources


def load_paq_entries_from_path(paq_path: Path) -> dict[str, bytes]:
    entries: dict[str, bytes] = {}
    if not paq_path.exists():
        raise FileNotFoundError(f"Missing PAQ archive: {paq_path}")
    for name, data in paq.iter_entries(paq_path):
        entries[name.replace("\\", "/")] = data
    return entries


def load_paq_entries(assets_dir: Path) -> dict[str, bytes]:
    return load_paq_entries_from_path(Path(assets_dir) / PAQ_NAME)


def _load_texture_from_bytes(data: bytes, fmt: str) -> rl.Texture:
    image = rl.load_image_from_memory(fmt, cast(str, data), len(data))
    texture = rl.load_texture_from_image(image)
    rl.unload_image(image)
    rl.set_texture_filter(texture, rl.TextureFilter.TEXTURE_FILTER_BILINEAR)
    return texture


def _apply_texture_settings(texture: rl.Texture, *, clamp: bool, point_filter: bool) -> None:
    if clamp:
        rl.set_texture_wrap(texture, rl.TextureWrap.TEXTURE_WRAP_CLAMP)
    if point_filter:
        rl.set_texture_filter(texture, rl.TextureFilter.TEXTURE_FILTER_POINT)


def _load_texture_asset_from_bytes(rel_path: str, data: bytes | None) -> rl.Texture | None:
    if data is None:
        raise FileNotFoundError(f"Missing asset data: {rel_path}")
    texture: rl.Texture | None
    if rel_path.lower().endswith(".jaz"):
        jaz_image = jaz.decode_jaz_bytes(data)
        buf = io.BytesIO()
        jaz_image.composite_image().save(buf, format="PNG")
        texture = _load_texture_from_bytes(buf.getvalue(), ".png")
    elif rel_path.lower().endswith(".tga"):
        img = Image.open(io.BytesIO(data))
        buf = io.BytesIO()
        img.save(buf, format="PNG")
        texture = _load_texture_from_bytes(buf.getvalue(), ".png")
    elif rel_path.lower().endswith((".jpg", ".jpeg")):
        img = Image.open(io.BytesIO(data))
        buf = io.BytesIO()
        img.save(buf, format="PNG")
        texture = _load_texture_from_bytes(buf.getvalue(), ".png")
    else:
        texture = None
    return texture


def _build_small_font(textures: dict[TextureId, rl.Texture], widths_data: bytes) -> SmallFontData:
    from grim.fonts.small import SmallFontData

    return SmallFontData(
        widths=list(widths_data),
        texture=textures[TextureId.SMALL_WHITE],
    )


def load_runtime_resources(assets_dir: Path) -> RuntimeResources:
    entries = load_paq_entries(Path(assets_dir))
    textures: dict[TextureId, rl.Texture] = {}
    for texture_id, spec in TEXTURE_SPECS.items():
        texture = _load_texture_asset_from_bytes(spec.rel_path, entries.get(spec.rel_path))
        if texture is None:
            raise FileNotFoundError(f"Missing runtime texture: {spec.rel_path}")
        _apply_texture_settings(texture, clamp=bool(spec.clamp), point_filter=bool(spec.point_filter))
        textures[texture_id] = texture

    widths_data = entries.get("load/smallFnt.dat")
    if widths_data is None:
        raise FileNotFoundError("Missing runtime font widths: load/smallFnt.dat")

    resources = RuntimeResources(
        assets_dir=Path(assets_dir),
        textures=textures,
        small_font=_build_small_font(textures, widths_data),
    )
    register_runtime_resources(resources)
    return resources


def unload_runtime_resources(resources: RuntimeResources | None) -> None:
    if resources is None:
        return
    unregister_runtime_resources(resources.assets_dir)
    resources.unload()



from __future__ import annotations

import datetime as dt
import faulthandler
import time
import webbrowser
from pathlib import Path

from grim import music
from grim.app import RunViewHooks, run_view
from grim.config import ensure_crimson_cfg
from grim.console import (
    CommandHandler,
    ConsoleState,
    create_console,
    register_boot_commands,
    register_core_cvars,
)
from grim.rand import Crand
from grim.raylib_api import rl
from grim.view import View

from ..assets_fetch import download_missing_paqs
from ..debug import set_debug_enabled
from ..demo_trial import (
    DEMO_QUEST_GRACE_TIME_MS,
    DEMO_TOTAL_PLAY_TIME_MS,
    demo_trial_overlay_info,
    format_demo_trial_time,
)
from ..game_modes import GameMode
from ..net.debug_log import close_lan_debug_log, init_lan_debug_log, lan_debug_log
from ..persistence.save_status import ensure_game_status
from ..render.rtx.mode import cycle_rtx_render_mode, mode_from_rtx_flag, parse_rtx_render_mode
from .loop_view import GameLoopView
from .types import GameConfig, GameState, LockstepEndpoint

CRIMSON_PAQ_NAME = "crimson.paq"
MUSIC_PAQ_NAME = "music.paq"
SFX_PAQ_NAME = "sfx.paq"
AUTOEXEC_NAME = "autoexec.txt"
REQUIRED_RUNTIME_PAQS: tuple[str, ...] = (CRIMSON_PAQ_NAME, MUSIC_PAQ_NAME, SFX_PAQ_NAME)


def _runtime_download_targets(assets_dir: Path) -> tuple[str, ...]:
    return tuple(name for name in REQUIRED_RUNTIME_PAQS if not (assets_dir / name).is_file())


def _require_runtime_assets(assets_dir: Path) -> None:
    missing = [name for name in REQUIRED_RUNTIME_PAQS if not (assets_dir / name).is_file()]
    if missing:
        joined = ", ".join(missing)
        raise FileNotFoundError(f"assets: missing required archives: {joined}")

def _parse_float_arg(value: str) -> float:
    try:
        return float(value)
    except ValueError:
        return 0.0


def _apply_debug_console_defaults(console: ConsoleState, *, debug: bool) -> None:
    if not bool(debug):
        return
    console.register_cvar("cv_showFPS", "1")


def _boot_command_handlers(state: GameState) -> dict[str, CommandHandler]:
    console = state.console

    def cmd_set_gamma_ramp(args: list[str]) -> None:
        if len(args) != 1:
            console.log.log("setGammaRamp <scalar > 0>")
            console.log.log("Command adjusts gamma ramp linearly by multiplying with given scalar")
            return
        value = _parse_float_arg(args[0])
        state.gamma_ramp = value
        console.log.log(f"Gamma ramp regenerated and multiplied with {value:.6f}")

    def cmd_snd_add_game_tune(args: list[str]) -> None:
        if len(args) != 1:
            console.log.log("snd_addGameTune <tuneName.ogg>")
            return
        audio = state.audio
        if audio is None:
            return
        rel_path = f"music/{args[0]}"
        result = music.load_music_track(audio.music, state.assets_dir, rel_path, console=console)
        if result is None:
            return
        track_key, _track_id = result
        music.queue_track(audio.music, track_key)

    def cmd_generate_terrain(_args: list[str]) -> None:
        state.terrain_regenerate_requested = True

    def cmd_tell_time_survived(_args: list[str]) -> None:
        seconds = int(max(0.0, float(state.survival_elapsed_ms)) * 0.00100000005)
        console.log.log(f"Survived: {seconds} seconds.")

    def cmd_set_resource_paq(args: list[str]) -> None:
        if len(args) != 1:
            console.log.log("setresourcepaq <resourcepaq>")
            return
        console.log.log("setresourcepaq is not supported in the rewrite.")

    def cmd_load_texture(args: list[str]) -> None:
        if len(args) != 1:
            console.log.log("loadtexture <texturefileid>")
            return
        console.log.log("loadtexture is not supported in the rewrite.")

    def cmd_open_url(args: list[str]) -> None:
        if len(args) != 1:
            console.log.log("openurl <url>")
            return
        url = args[0]
        ok = False
        try:
            ok = webbrowser.open(url)
        except (OSError, webbrowser.Error):
            ok = False
        if ok:
            console.log.log(f"Launching web browser ({url})..")
        else:
            console.log.log("Failed to launch web browser.")

    def cmd_snd_freq_adjustment(_args: list[str]) -> None:
        state.snd_freq_adjustment_enabled = not state.snd_freq_adjustment_enabled
        if state.snd_freq_adjustment_enabled:
            console.log.log("Sound frequency adjustment is now enabled.")
        else:
            console.log.log("Sound frequency adjustment is now disabled.")

    def cmd_demo_trial_set_playtime(args: list[str]) -> None:
        if len(args) != 1:
            console.log.log("demoTrialSetPlaytime <ms>")
            return
        try:
            value = int(float(args[0]))
        except ValueError:
            value = 0
        state.status.game_sequence_id = max(0, value)
        state.status.save_if_dirty()
        console.log.log(f"demo trial: playtime={state.status.game_sequence_id}ms (total {DEMO_TOTAL_PLAY_TIME_MS}ms)")

    def cmd_demo_trial_set_grace(args: list[str]) -> None:
        if len(args) != 1:
            console.log.log("demoTrialSetGrace <ms>")
            return
        try:
            value = int(float(args[0]))
        except ValueError:
            value = 0
        state.demo_trial_elapsed_ms = max(0, value)
        console.log.log(f"demo trial: quest grace={state.demo_trial_elapsed_ms}ms (total {DEMO_QUEST_GRACE_TIME_MS}ms)")

    def cmd_demo_trial_reset(_args: list[str]) -> None:
        state.status.game_sequence_id = 0
        state.status.save_if_dirty()
        state.demo_trial_elapsed_ms = 0
        console.log.log("demo trial: timers reset")

    def cmd_demo_trial_info(_args: list[str]) -> None:
        mode_raw = state.config.gameplay.mode
        try:
            mode_id = GameMode(mode_raw)
        except ValueError:
            mode_id = GameMode.DEMO
        quest_level = None
        match mode_id:
            case GameMode.QUESTS:
                quest_level = state.pending_quest_level
            case _:
                pass
        info = demo_trial_overlay_info(
            demo_build=bool(state.demo_enabled),
            game_mode_id=mode_id,
            global_playtime_ms=int(state.status.game_sequence_id),
            quest_grace_elapsed_ms=int(state.demo_trial_elapsed_ms),
            quest_level=quest_level,
        )
        remaining = format_demo_trial_time(info.remaining_ms)
        console.log.log(
            "demo trial: "
            f"demo={int(state.demo_enabled)} "
            f"mode={int(mode_id)} "
            f"quest={(quest_level.text if quest_level is not None else '0.0')} "
            f"playtime={int(state.status.game_sequence_id)}ms "
            f"grace={int(state.demo_trial_elapsed_ms)}ms "
            f"visible={int(info.visible)} "
            f"kind={info.kind} "
            f"remaining={remaining}",
        )

    def cmd_render_mode(args: list[str]) -> None:
        if len(args) > 1:
            console.log.log("rendermode <classic|rtx>")
            return
        if not args:
            console.log.log(f"Render mode is '{state.rtx_mode.value}'.")
            return
        try:
            mode = parse_rtx_render_mode(args[0])
        except ValueError:
            console.log.log("rendermode <classic|rtx>")
            return
        state.rtx_mode = mode
        console.log.log(f"Render mode set to '{state.rtx_mode.value}'.")

    def cmd_toggle_rtx(args: list[str]) -> None:
        if args:
            console.log.log("togglertx")
            return
        state.rtx_mode = cycle_rtx_render_mode(state.rtx_mode)
        console.log.log(f"Render mode set to '{state.rtx_mode.value}'.")

    return {
        "setGammaRamp": cmd_set_gamma_ramp,
        "snd_addGameTune": cmd_snd_add_game_tune,
        "generateterrain": cmd_generate_terrain,
        "telltimesurvived": cmd_tell_time_survived,
        "setresourcepaq": cmd_set_resource_paq,
        "loadtexture": cmd_load_texture,
        "openurl": cmd_open_url,
        "sndfreqadjustment": cmd_snd_freq_adjustment,
        "demoTrialSetPlaytime": cmd_demo_trial_set_playtime,
        "demoTrialSetGrace": cmd_demo_trial_set_grace,
        "demoTrialReset": cmd_demo_trial_reset,
        "demoTrialInfo": cmd_demo_trial_info,
        "rendermode": cmd_render_mode,
        "togglertx": cmd_toggle_rtx,
    }


def _resolve_assets_dir(config: GameConfig) -> Path:
    if config.assets_dir is not None:
        return config.assets_dir
    return config.base_dir


def run_game(config: GameConfig) -> None:
    if config.debug:
        set_debug_enabled(True)
    base_dir = config.base_dir
    base_dir.mkdir(parents=True, exist_ok=True)
    crash_path = base_dir / "crash.log"
    crash_file = crash_path.open("a", encoding="utf-8", buffering=1)
    faulthandler.enable(crash_file)
    crash_file.write(f"\n[{dt.datetime.now().isoformat()}] run_game start\n")
    cfg = ensure_crimson_cfg(base_dir)
    width = cfg.display.width if config.width is None else config.width
    height = cfg.display.height if config.height is None else config.height
    rng = Crand(config.seed)
    assets_dir = _resolve_assets_dir(config)
    console = create_console(base_dir, assets_dir=assets_dir)
    status = ensure_game_status(base_dir)
    state: GameState | None = None
    try:
        state = GameState(
            base_dir=base_dir,
            assets_dir=assets_dir,
            rng=rng,
            config=cfg,
            status=status,
            console=console,
            demo_enabled=config.demo_enabled,
            preserve_bugs=config.preserve_bugs,
            skip_intro=config.no_intro,
            resources=None,
            audio=None,
            session_start=time.monotonic(),
            rtx_mode=mode_from_rtx_flag(bool(config.rtx)),
            pending_network_session=config.pending_network_session,
        )
        pending = config.pending_network_session
        if pending is not None:
            from ..net.lockstep_protocol import current_build_id

            endpoint = pending.config.endpoint
            if isinstance(endpoint, LockstepEndpoint):
                host = str(endpoint.host)
                port = int(endpoint.port)
            else:
                host = str(endpoint.relay_host)
                port = int(endpoint.relay_port)
            log_path = init_lan_debug_log(
                base_dir=base_dir,
                role=str(pending.role),
                mode=str(pending.config.mode),
                build_id=str(current_build_id()),
                host=host,
                port=int(port),
                player_count=pending.config.player_count,
                auto_start=pending.auto_start,
                debug_enabled=config.debug,
            )
            lan_debug_log(
                "run_game_session",
                width=int(width),
                height=int(height),
                fps=int(config.fps),
                preserve_bugs=config.preserve_bugs,
            )
            console.log.log(f"lan debug log: {log_path}")
            print(f"[lan-debug] role={pending.role} log={log_path}")
        register_boot_commands(console, _boot_command_handlers(state))
        register_core_cvars(console, width, height)
        _apply_debug_console_defaults(console, debug=config.debug)
        console.log.log("crimson: boot start")
        console.log.log(f"config: {cfg.display.width}x{cfg.display.height} windowed={cfg.display.windowed}")
        console.log.log(f"status: {status.path.name} loaded")
        console.log.log(f"assets: {assets_dir}")
        download_missing_paqs(assets_dir, console, names=_runtime_download_targets(assets_dir))
        _require_runtime_assets(assets_dir)
        console.log.log(f"assets: required archives ready ({', '.join(REQUIRED_RUNTIME_PAQS)})")
        console.log.log(f"commands: {len(console.commands)} registered")
        console.log.log(f"cvars: {len(console.cvars)} registered")
        console.exec_line("exec autoexec.txt")
        console.log.flush()
        config_flags = 0
        if not cfg.display.windowed:
            config_flags |= rl.ConfigFlags.FLAG_FULLSCREEN_MODE
        view: View = GameLoopView(state)
        run_view(
            view,
            width=width,
            height=height,
            title="Crimsonland",
            fps=config.fps,
            config_flags=config_flags,
            exit_key=rl.KeyboardKey.KEY_NULL,
            hooks=RunViewHooks(
                should_close=view.should_close,
                consume_screenshot_request=view.consume_screenshot_request,
            ),
        )
        if state is not None:
            state.status.save_if_dirty()
    finally:
        close_lan_debug_log()
        faulthandler.disable()
        crash_file.close()



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

import msgspec

from ..types import Projectile, ProjectileTemplateId
from .behaviors import (
    _linger_default,
    _linger_gauss_gun,
    _linger_ion_cannon,
    _linger_ion_minigun,
    _linger_ion_rifle,
    _post_hit_ion_common,
    _post_hit_ion_rifle,
    _post_hit_plague_spreader,
    _post_hit_plasma_cannon,
    _post_hit_pulse_gun,
    _post_hit_shrinkifier,
    _pre_hit_splitter,
)

if TYPE_CHECKING:
    from .behaviors import _ProjectileHitInfo, _ProjectileUpdateCtx


LingerHandler = Callable[["_ProjectileUpdateCtx", Projectile], None]
PreHitHandler = Callable[["_ProjectileUpdateCtx", Projectile, int], None]
PostHitHandler = Callable[["_ProjectileUpdateCtx", "_ProjectileHitInfo"], None]


def _pre_hit_none(_ctx: _ProjectileUpdateCtx, _proj: Projectile, _hit_idx: int) -> None:
    return


def _post_hit_none(_ctx: _ProjectileUpdateCtx, _hit: _ProjectileHitInfo) -> None:
    return


class PrimaryProjectileRule(msgspec.Struct, frozen=True):
    linger: LingerHandler
    pre_hit: PreHitHandler = _pre_hit_none
    post_hit: PostHitHandler = _post_hit_none
    stop_on_hit: bool = True
    emit_default_freeze_shard: bool = True
    reset_shock_chain_on_linger: bool = False


_DEFAULT_RULE = PrimaryProjectileRule(linger=_linger_default)


PRIMARY_PROJECTILE_RULE_BY_TYPE_ID: dict[ProjectileTemplateId, PrimaryProjectileRule] = {
    ProjectileTemplateId.GAUSS_GUN: PrimaryProjectileRule(
        linger=_linger_gauss_gun,
        stop_on_hit=False,
        emit_default_freeze_shard=False,
    ),
    ProjectileTemplateId.FIRE_BULLETS: PrimaryProjectileRule(
        linger=_linger_default,
        stop_on_hit=False,
        emit_default_freeze_shard=False,
    ),
    ProjectileTemplateId.BLADE_GUN: PrimaryProjectileRule(
        linger=_linger_default,
        stop_on_hit=False,
    ),
    ProjectileTemplateId.PULSE_GUN: PrimaryProjectileRule(
        linger=_linger_default,
        post_hit=_post_hit_pulse_gun,
    ),
    ProjectileTemplateId.ION_RIFLE: PrimaryProjectileRule(
        linger=_linger_ion_rifle,
        post_hit=_post_hit_ion_rifle,
        reset_shock_chain_on_linger=True,
    ),
    ProjectileTemplateId.ION_MINIGUN: PrimaryProjectileRule(
        linger=_linger_ion_minigun,
        post_hit=_post_hit_ion_common,
        reset_shock_chain_on_linger=True,
    ),
    ProjectileTemplateId.ION_CANNON: PrimaryProjectileRule(
        linger=_linger_ion_cannon,
        post_hit=_post_hit_ion_common,
    ),
    ProjectileTemplateId.SHRINKIFIER: PrimaryProjectileRule(
        linger=_linger_default,
        post_hit=_post_hit_shrinkifier,
    ),
    ProjectileTemplateId.PLASMA_CANNON: PrimaryProjectileRule(
        linger=_linger_default,
        post_hit=_post_hit_plasma_cannon,
    ),
    ProjectileTemplateId.SPLITTER_GUN: PrimaryProjectileRule(
        linger=_linger_default,
        pre_hit=_pre_hit_splitter,
    ),
    ProjectileTemplateId.PLAGUE_SPREADER: PrimaryProjectileRule(
        linger=_linger_default,
        post_hit=_post_hit_plague_spreader,
    ),
}


def primary_rule_for_type_id(type_id: ProjectileTemplateId) -> PrimaryProjectileRule:
    return PRIMARY_PROJECTILE_RULE_BY_TYPE_ID.get(type_id, _DEFAULT_RULE)


__all__ = [
    "PRIMARY_PROJECTILE_RULE_BY_TYPE_ID",
    "PrimaryProjectileRule",
    "primary_rule_for_type_id",
]
